"""Initialization of file-keepers extensions."""

from __future__ import annotations

import logging
import os

from pluggy import HookimplMarker, PluginManager

from file_keeper.core import storage, upload, utils

from . import spec

hookimpl = HookimplMarker(spec.name)
plugin = PluginManager(spec.name)
plugin.add_hookspecs(spec)
log = logging.getLogger(__name__)


@utils.run_once
def setup():
    """Discover and register file-keeper extensions."""
    try:
        plugin.load_setuptools_entrypoints(spec.name)
    except ModuleNotFoundError:
        msg = (
            "File-keeper extension is missing, try to re-install the problematic package."
            + " Only default adapters will be loaded"
        )
        log.exception(msg)
        plugin.load_setuptools_entrypoints(spec.name, "default")
    for name in os.getenv("FILE_KEEPER_DISABLED_EXTENSIONS", "").split():
        undesired = plugin.get_plugin(name)
        if plugin.is_registered(undesired):
            plugin.unregister(undesired)

    register(True)


def register(reset: bool = False):
    """Register built-in units."""
    if reset:
        storage.location_transformers.reset()
        upload.upload_factories.reset()
        storage.adapters.reset()
    plugin.hook.register_location_transformers(registry=storage.location_transformers)
    plugin.hook.register_upload_factories(registry=upload.upload_factories)
    plugin.hook.register_adapters(registry=storage.adapters)


if os.getenv("FILE_KEEPER_AUTO_SETUP"):
    setup()



"""Plugin specification."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pluggy import HookspecMarker

if TYPE_CHECKING:
    from file_keeper.core import storage, types, upload
    from file_keeper.core.registry import Registry


name = "file_keeper_ext"


hookspec = HookspecMarker(name)


@hookspec
def register_adapters(registry: Registry[type[storage.Storage]]):
    """Register storage adapters."""


@hookspec
def register_upload_factories(registry: Registry[upload.UploadFactory, type]):
    """Register upload factories.

    Example:
        ```py
        def register_upload_factories(registry):
            def string_into_upload(value: str):
                return bytes(value, "utf8")
            registry.register(str, string_into_upload)
        ```
    """


@hookspec
def register_location_transformers(registry: Registry[types.LocationTransformer]):
    """Register location transformers.

    Example:
        ```py
        def register_location_transformers(registry):
            def lower(location, upload, extras):
                return location.lower().

            registry.register("lower", lower)
        ```
    """



"""Default implementations of file-keeper units."""

from __future__ import annotations

import contextlib
import io
import logging
import mimetypes
import tempfile
from typing import cast

import magic

from file_keeper import Registry, Storage, Upload, ext, types
from file_keeper.core.upload import UploadFactory

from . import adapters, transformers

SAMPLE_SIZE = 1024 * 2

log = logging.getLogger(__name__)


@ext.hookimpl
def register_location_transformers(registry: Registry[types.LocationTransformer]):
    """Built-in location transformers."""
    registry.register("datetime_prefix", transformers.datetime_prefix_transformer)
    registry.register("datetime_with_extension", transformers.datetime_with_extension_transformer)
    registry.register("fix_extension", transformers.fix_extension_transformer)
    registry.register("safe_relative_path", transformers.safe_relative_path_transformer)
    registry.register("uuid", transformers.uuid_transformer)
    registry.register("uuid_prefix", transformers.uuid_prefix_transformer)
    registry.register("uuid_with_extension", transformers.uuid_with_extension_transformer)
    registry.register("static_uuid", transformers.static_uuid_transformer)


@ext.hookimpl
def register_upload_factories(registry: Registry[UploadFactory, type]):
    """Built-in upload converter."""
    registry.register(tempfile.SpooledTemporaryFile, tempfile_into_upload)
    registry.register(io.TextIOWrapper, textiowrapper_into_upload)


with contextlib.suppress(ImportError):
    import cgi

    @ext.hookimpl(specname="register_upload_factories")
    def _(registry: Registry[UploadFactory, type]):
        registry.register(cgi.FieldStorage, cgi_field_storage_into_upload)

    def cgi_field_storage_into_upload(value: cgi.FieldStorage):
        """cgi.field-into-upload factory."""
        if not value.filename or not value.file:
            return None

        mime, _encoding = mimetypes.guess_type(value.filename)
        if not mime:
            mime = magic.from_buffer(value.file.read(SAMPLE_SIZE), True)
            _ = value.file.seek(0)

        _ = value.file.seek(0, 2)
        size = value.file.tell()
        _ = value.file.seek(0)

        return Upload(
            value.file,
            value.filename,
            size,
            mime,
        )


with contextlib.suppress(ImportError):
    from werkzeug.datastructures import FileStorage

    @ext.hookimpl(specname="register_upload_factories")
    def _(registry: Registry[UploadFactory, type]):
        registry.register(FileStorage, werkzeug_file_storage_into_upload)

    def werkzeug_file_storage_into_upload(value: FileStorage):
        """werkzeug.FileStorage-into-upload converter."""
        name: str = value.filename or value.name or ""
        if value.content_length:
            size = value.content_length
        else:
            _ = value.stream.seek(0, 2)
            size = value.stream.tell()
            _ = value.stream.seek(0)

        mime = magic.from_buffer(value.stream.read(SAMPLE_SIZE), True)
        _ = value.stream.seek(0)

        return Upload(value.stream, name, size, mime)


def tempfile_into_upload(value: tempfile.SpooledTemporaryFile[bytes]):
    """tmpfile-into-upload converter."""
    mime = magic.from_buffer(value.read(SAMPLE_SIZE), True)
    _ = value.seek(0, 2)
    size = value.tell()
    _ = value.seek(0)

    return Upload(value, value.name or "", size, mime)


def textiowrapper_into_upload(value: io.TextIOWrapper):
    """TextIO-into-upload converter."""
    return cast(io.BufferedReader, value.buffer)


# --8<-- [start:register]
@ext.hookimpl
def register_adapters(registry: Registry[type[Storage]]):  # noqa: C901
    """Built-in storage adapters."""
    registry.register("file_keeper:fs", adapters.FsStorage)
    # --8<-- [end:register]
    registry.register("file_keeper:null", adapters.NullStorage)
    registry.register("file_keeper:memory", adapters.MemoryStorage)
    registry.register("file_keeper:zip", adapters.ZipStorage)

    if adapters.RedisStorage:
        registry.register("file_keeper:redis", adapters.RedisStorage)

    if adapters.OpenDalStorage:
        registry.register("file_keeper:opendal", adapters.OpenDalStorage)

    if adapters.LibCloudStorage:
        registry.register("file_keeper:libcloud", adapters.LibCloudStorage)

    if adapters.GoogleCloudStorage:
        registry.register("file_keeper:gcs", adapters.GoogleCloudStorage)

    if adapters.S3Storage:
        registry.register("file_keeper:s3", adapters.S3Storage)

    if adapters.FilebinStorage:
        registry.register("file_keeper:filebin", adapters.FilebinStorage)

    if adapters.SqlAlchemyStorage:
        registry.register("file_keeper:sqlalchemy", adapters.SqlAlchemyStorage)

    if adapters.AzureBlobStorage:
        registry.register("file_keeper:azure_blob", adapters.AzureBlobStorage)

    if adapters.ObjectStoreStorage:
        registry.register("file_keeper:object_store", adapters.ObjectStoreStorage)

    if adapters.FsSpecStorage:
        registry.register("file_keeper:fsspec", adapters.FsSpecStorage)

    registry.register("file_keeper:proxy", adapters.ProxyStorage)  # pyright: ignore[reportArgumentType]



"""Exception definitions for the extension.

Hierarchy:

* Exception
    * [FilesError][file_keeper.exc.FilesError]
        * [StorageError][file_keeper.exc.StorageError]
            * [UnknownAdapterError][file_keeper.exc.UnknownAdapterError]
            * [UnknownStorageError][file_keeper.exc.UnknownStorageError]
            * [UnsupportedOperationError][file_keeper.exc.UnsupportedOperationError]
            * [PermissionError][file_keeper.exc.PermissionError]
            * [LocationError][file_keeper.exc.LocationError]
                * [MissingFileError][file_keeper.exc.MissingFileError]
                * [ExistingFileError][file_keeper.exc.ExistingFileError]
            * [StorageDataError][file_keeper.exc.StorageDataError]
            * [ExtrasError][file_keeper.exc.ExtrasError]
                * [MissingExtrasError][file_keeper.exc.MissingExtrasError]
            * [InvalidStorageConfigurationError][file_keeper.exc.InvalidStorageConfigurationError]
                * [MissingStorageConfigurationError][file_keeper.exc.MissingStorageConfigurationError]
            * [UploadError][file_keeper.exc.UploadError]
                * [WrongUploadTypeError][file_keeper.exc.WrongUploadTypeError]
                * [LocationTransformerError][file_keeper.exc.LocationTransformerError]
                * [ContentError][file_keeper.exc.ContentError]
                * [LargeUploadError][file_keeper.exc.LargeUploadError]
                    * [UploadOutOfBoundError][file_keeper.exc.UploadOutOfBoundError]
                * [UploadMismatchError][file_keeper.exc.UploadMismatchError]
                    * [UploadTypeMismatchError][file_keeper.exc.UploadTypeMismatchError]
                    * [UploadHashMismatchError][file_keeper.exc.UploadHashMismatchError]
                    * [UploadSizeMismatchError][file_keeper.exc.UploadSizeMismatchError]
                * [ResumableUploadError][file_keeper.exc.ResumableUploadError]
                * [MultipartUploadError][file_keeper.exc.MultipartUploadError]


"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

Storage = Any


class FilesError(Exception):
    """Base error for catch-all scenario."""


class StorageError(FilesError):
    """Error related to storage."""


class UnknownStorageError(StorageError):
    """Storage with the given name is not configured."""

    def __init__(self, storage: str):
        super().__init__(f"Storage {storage} is not configured")


class UnknownAdapterError(StorageError):
    """Specified storage adapter is not registered."""

    def __init__(self, adapter: str):
        super().__init__(f"Storage adapter {adapter} is not registered")


class UnsupportedOperationError(StorageError):
    """Requested operation is not supported by storage."""

    def __init__(self, operation: str, storage: Storage):
        super().__init__(f"Operation {operation} is not supported by {storage} storage")


class InvalidStorageConfigurationError(StorageError):
    """Storage cannot be initialized with given configuration."""

    def __init__(self, adapter_or_storage: type | str, problem: str):
        if isinstance(adapter_or_storage, str):
            what = f"storage {adapter_or_storage}"
        else:
            what = f"storage adapter {adapter_or_storage.__name__}"

        super().__init__(f"Cannot initialize {what}: {problem}")


class PermissionError(StorageError):
    """Storage client does not have required permissions."""

    def __init__(self, storage: Storage, operation: str, problem: str):
        msg = f"Storage {storage} is not allowed to perform {operation} operation: {problem}"
        super().__init__(msg)


class MissingStorageConfigurationError(InvalidStorageConfigurationError):
    """Storage cannot be initialized due to missing option."""

    def __init__(self, adapter_or_storage: type | str, option: str):
        super().__init__(
            adapter_or_storage,
            f"{option} option is required",
        )


class LocationError(StorageError):
    """Storage cannot use given location."""

    tpl: str = "{storage} cannot use location {location}"

    def __init__(self, storage: Storage, location: str):
        super().__init__(self.tpl.format(storage=storage, location=location))


class MissingFileError(LocationError):
    """File does not exist."""

    tpl: str = "File {location} does not exist inside storage {storage}"


class ExistingFileError(LocationError):
    """File already exists."""

    tpl: str = "File {location} already exists inside storage {storage}"


class UploadError(StorageError):
    """Error related to file upload process."""


class MultipartUploadError(UploadError):
    """Error related to multipart upload process."""


class ResumableUploadError(UploadError):
    """Error related to resumable upload process."""


class LargeUploadError(UploadError):
    """Planned upload exceeds allowed size."""

    tpl: str = "Upload size {actual_size} surpasses max allowed size {max_size}"

    def __init__(self, actual_size: int, max_size: int):
        super().__init__(self.tpl.format(actual_size=actual_size, max_size=max_size))


class UploadOutOfBoundError(LargeUploadError):
    """Ongoing upload exceeds expected size."""

    tpl: str = "Upload size {actual_size} exceeds expected size {max_size}"


class UploadMismatchError(UploadError):
    """Expected value of file attribute doesn't match the actual value."""

    value_formatter: Callable[[Any], Any] = str

    def __init__(self, attribute: str, actual: Any, expected: Any):
        actual = self.value_formatter(actual)
        expected = self.value_formatter(expected)

        super().__init__(f"Actual value of {attribute}({actual})" + f" does not match expected value({expected})")


class UploadTypeMismatchError(UploadMismatchError):
    """Expected value of content type doesn't match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("content type", actual, expected)


class UploadHashMismatchError(UploadMismatchError):
    """Expected value of hash match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("content hash", actual, expected)


class UploadSizeMismatchError(UploadMismatchError):
    """Expected value of upload size doesn't match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("upload size", actual, expected)


class WrongUploadTypeError(UploadError):
    """Storage does not support given MIMEType."""

    def __init__(self, content_type: str):
        super().__init__(f"Type {content_type} is not supported by storage")


class LocationTransformerError(UploadError):
    """Undefined location transformer."""

    def __init__(self, transformer: str):
        super().__init__(f"Unknown location transformer {transformer}")


class ExtrasError(StorageError):
    """Wrong extras passed to storage method."""

    tpl: str = "Wrong extras: {problem}"

    def __init__(self, problem: Any):
        super().__init__(self.tpl.format(problem=problem))


class MissingExtrasError(ExtrasError):
    """Expected extras are missing from the call."""

    def __init__(self, key: Any):
        super().__init__(f"key {key} is missing")


class ContentError(UploadError):
    """Storage cannot accept uploaded content."""

    def __init__(self, storage: Storage, msg: str):
        super().__init__(f"{storage} rejected upload: {msg}")


class StorageDataError(StorageError):
    """Wrong [storage_data][file_keeper.FileData] used for operation."""

    tpl: str = "Wrong storage data: {problem}"

    def __init__(self, problem: Any):
        super().__init__(self.tpl.format(problem=problem))



"""SQLAlchemy adapter."""

from __future__ import annotations

import dataclasses
import os
from collections.abc import Iterable
from typing import Any

import sqlalchemy as sa
from typing_extensions import override

import file_keeper as fk


@dataclasses.dataclass()
class Settings(fk.Settings):
    """SQLAlchemy settings."""

    db_url: str = ""
    """URL of the storage DB."""
    table_name: str = ""
    """Name of the storage table."""
    location_column: str = ""
    """Name of the column that contains file location."""
    content_column: str = ""
    """Name of the column that contains file content."""

    engine: sa.engine.Engine = None  # pyright: ignore[reportAssignmentType]
    """Existing DB engine."""
    table: sa.Table = None  # pyright: ignore[reportAssignmentType]
    """Existing DB table."""
    location: sa.Column[str] = None  # pyright: ignore[reportAssignmentType]
    """Existing column for location."""
    content: sa.Column[bytes] = None  # pyright: ignore[reportAssignmentType]
    """Existing column for content."""

    def __post_init__(self, **kwargs: Any):  # noqa: C901, PLR0912
        super().__post_init__(**kwargs)

        if not self.engine:
            if not self.db_url:
                raise fk.exc.MissingStorageConfigurationError(self.name, "db_url")
            self.engine = sa.create_engine(self.db_url)

        if self.location is None:  # pyright: ignore[reportUnnecessaryComparison]
            if not self.location_column:
                raise fk.exc.MissingStorageConfigurationError(self.name, "location_column")
            self.location = sa.Column(self.location_column, sa.Text, primary_key=True)
        else:
            self.location_column = self.location.name

        if self.content is None:  # pyright: ignore[reportUnnecessaryComparison]
            if not self.content_column:
                raise fk.exc.MissingStorageConfigurationError(self.name, "content_column")
            self.content = sa.Column(self.content_column, sa.LargeBinary)
        else:
            self.content_column = self.content.name

        if self.table is None:  # pyright: ignore[reportUnnecessaryComparison]
            if not self.table_name:
                raise fk.exc.MissingStorageConfigurationError(self.name, "table_name")

            self.table = sa.Table(
                self.table_name,
                sa.MetaData(),
                self.location,
                self.content,
            )
        else:
            self.table_name = self.table.name

        inspector = sa.inspect(self.engine)
        if not inspector.has_table(self.table.name):
            if self.initialize:
                self.table.create(self.engine)
            else:
                raise fk.exc.InvalidStorageConfigurationError(self.name, f"table {self.table.name} does not exist")


class Reader(fk.Reader):
    """SQLAlchemy reader."""

    storage: SqlAlchemyStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        filepath = self.storage.full_path(data.location)

        stmt = (
            sa.select(self.storage.settings.content)
            .select_from(self.storage.settings.table)
            .where(self.storage.settings.location == filepath)
        )

        with self.storage.settings.engine.connect() as conn:
            row = conn.execute(stmt).fetchone()

        if row is None:
            raise fk.exc.MissingFileError(self, data.location)

        return row


class Uploader(fk.Uploader):
    """SQLAlchemy uploader."""

    storage: SqlAlchemyStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        filepath = self.storage.full_path(location)
        reader = upload.hashing_reader()

        values: dict[Any, Any] = {
            self.storage.settings.location: filepath,
            self.storage.settings.content: reader.read(),
        }

        table = self.storage.settings.table

        with self.storage.settings.engine.begin() as conn:
            if conn.scalar(sa.select(1).select_from(table).where(self.storage.settings.location == filepath)):
                if self.storage.settings.override_existing:
                    stmt = sa.update(table).where(self.storage.settings.location == filepath).values(values)
                    conn.execute(stmt)
                else:
                    raise fk.exc.ExistingFileError(self.storage, location)
            else:
                stmt = sa.insert(self.storage.settings.table).values(values)
                conn.execute(stmt)

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            reader.get_hash(),
        )


class Manager(fk.Manager):
    """SQLAlchemy manager."""

    storage: SqlAlchemyStorage
    capabilities: fk.Capability = (
        fk.Capability.SCAN
        | fk.Capability.REMOVE
        | fk.Capability.EXISTS
        | fk.Capability.ANALYZE
        | fk.Capability.COPY
        | fk.Capability.MOVE
    )

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        srcpath = self.storage.full_path(data.location)
        destpath = self.storage.full_path(location)

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if self.exists(fk.FileData(location), extras):
            if self.storage.settings.override_existing:
                self.remove(fk.FileData(location), extras)
            else:
                raise fk.exc.ExistingFileError(self.storage, location)

        stmt = (
            sa.update(self.storage.settings.table)
            .where(self.storage.settings.location == srcpath)
            .values({self.storage.settings.location: destpath})
        )
        with self.storage.settings.engine.begin() as conn:
            conn.execute(stmt)

        return fk.FileData.from_object(data, location=location)

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        srcpath = self.storage.full_path(data.location)
        destpath = self.storage.full_path(location)

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if self.exists(fk.FileData(location), extras):
            if self.storage.settings.override_existing:
                self.remove(fk.FileData(location), extras)
            else:
                raise fk.exc.ExistingFileError(self.storage, location)

        with self.storage.settings.engine.begin() as conn:
            content = conn.scalar(
                sa.select(self.storage.settings.content).where(self.storage.settings.location == srcpath)
            )

            conn.execute(
                sa.insert(self.storage.settings.table).values(
                    {
                        self.storage.settings.location: destpath,
                        self.storage.settings.content: content,
                    }
                )
            )

        return fk.FileData.from_object(data, location=location)

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)

        stmt = sa.select(self.storage.settings.content).where(self.storage.settings.location == filepath)
        with self.storage.settings.engine.connect() as conn:
            content = conn.scalar(stmt)

        if not content:
            raise fk.exc.MissingFileError(self.storage, location)

        upload = fk.make_upload(content)
        reader = upload.hashing_reader()
        reader.exhaust()
        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            reader.get_hash(),
        )

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        stmt = sa.select(1).select_from(self.storage.settings.table).where(self.storage.settings.location == filepath)
        with self.storage.settings.engine.connect() as conn:
            return bool(conn.scalar(stmt))

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        stmt = sa.select(self.storage.settings.location).select_from(self.storage.settings.table)
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        with self.storage.settings.engine.connect() as conn:
            for row in conn.execute(stmt):
                if row[0].startswith(path):
                    yield os.path.relpath(row[0], path)

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        if not self.exists(data, extras):
            return False

        stmt = sa.delete(self.storage.settings.table).where(
            self.storage.settings.location == filepath,
        )
        with self.storage.settings.engine.begin() as conn:
            conn.execute(stmt)
        return True


class SqlAlchemyStorage(fk.Storage):
    """SQLAlchemy storage.

    This storage uses a SQL database to store files as BLOBs. It requires SQLAlchemy
    to be installed.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:sqlalchemy",
        "db_url": "sqlite:///file_keeper.db",
        "table_name": "files",
        "location_column": "location",
        "content_column": "content",
        "override_existing": True,
        "initialize": True,
    }

    storage = fk.make_storage("sqlalchemy", settings)
    ```

    Note:
    * The `db_url` setting specifies the database connection URL.
    * The `table_name`, `location_column`, and `content_column` settings specify
      the table and column names to use for storing files.
    * If the specified table does not exist and `initialize` is set to `True`,
        it will be created automatically.
    * If `override_existing` is set to `True`, existing files will be overwritten
        during upload, copy, or move operations.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import dataclasses
import fnmatch
import functools
import inspect
import json
import logging
import os
import pathlib
from abc import ABC
from collections.abc import Callable, Iterable, Mapping
from typing import Any, ClassVar, Literal, TypeAlias, cast

from typing_extensions import ParamSpec, TypeVar, override

from . import data, exceptions, types, utils
from .registry import Registry
from .upload import Upload, make_upload

try:
    from platformdirs import user_config_dir  # pyright: ignore[reportAssignmentType]
except ImportError:

    def user_config_dir(
        appname: str | None = None,
        appauthor: str | Literal[False] | None = None,
    ):
        """Mock for user config locator."""
        return


try:
    import tomllib as toml  # pyright: ignore[reportMissingImports]
except ImportError:
    try:
        import tomli as toml
    except ImportError:
        toml = None

try:
    import yaml
except ImportError:
    yaml = None


P = ParamSpec("P")
T = TypeVar("T")
S = TypeVar("S", bound="Storage")
TCallable = TypeVar("TCallable", bound=Callable[..., Any])

log = logging.getLogger(__name__)

Capability: TypeAlias = utils.Capability

adapters = Registry["type[Storage]"]()
storages = Registry["Storage"]()

location_transformers = Registry[types.LocationTransformer]()


def requires_capability(capability: Capability):
    """Decorator to ensure a method requires a specific capability."""

    def decorator(func: TCallable) -> TCallable:
        """Wraps the function to check for the required capability."""

        @functools.wraps(func)
        def method(self: Any, *args: Any, **kwargs: Any):
            """Executes the function if the capability is supported."""
            if not self.supports(capability):
                raise exceptions.UnsupportedOperationError(str(capability.name), self)
            return func(self, *args, **kwargs)

        return cast(Any, method)

    return decorator


class StorageService:
    """Base class for services used by storage.

    StorageService.capabilities reflect all operations provided by the
    service.

    Example:
        ```py
        class Uploader(StorageService):
            capabilities = Capability.CREATE
        ```
    """

    capabilities: Capability = Capability.NONE

    def __init__(self, storage: Storage):
        self.storage = storage


class Uploader(StorageService):
    """Service responsible for writing data into a storage.

    `Storage` internally calls methods of this service. For example,
    `Storage.upload(location, upload, **kwargs)` results in
    `Uploader.upload(location, upload, kwargs)`.

    Example:
        ```python
        class MyUploader(Uploader):
            capabilities = Capability.CREATE

            def upload(
                self, location: types.Location, upload: Upload, extras: dict[str, Any]
            ) -> FileData:
                reader = upload.hashing_reader()

                with open(location, "wb") as dest:
                    dest.write(reader.read())

                return FileData(
                    location, upload.size,
                    upload.content_type,
                    reader.get_hash()
                )
        ```
    """

    def upload(self, location: types.Location, upload: Upload, extras: dict[str, Any]) -> data.FileData:
        """Upload file using single stream.

        Args:
            location: The destination location for the upload.
            upload: The Upload object containing the file data.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def resumable_start(self, location: types.Location, size: int, extras: dict[str, Any]) -> data.FileData:
        """Prepare everything for resumable upload.

        Args:
            location: The destination location for the upload.
            size: Expected upload size.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def resumable_refresh(self, data: data.FileData, extras: dict[str, Any]) -> data.FileData:
        """Show details of the incomplete resumable upload.

        Args:
            data: The FileData object containing the upload metadata.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def resumable_resume(self, data: data.FileData, upload: Upload, extras: dict[str, Any]) -> data.FileData:
        """Resume the interrupted resumable upload.

        Args:
            data: The FileData object containing the upload metadata.
            upload: The Upload object containing the content.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def resumable_remove(self, data: data.FileData, extras: dict[str, Any]) -> bool:
        """Remove incomplete resumable upload.

        Args:
            data: The FileData object containing the upload metadata.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def multipart_start(self, location: types.Location, size: int, extras: dict[str, Any]) -> data.FileData:
        """Prepare everything for multipart(resumable) upload.

        Args:
            location: The destination location for the upload.
            size: Expected upload size.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def multipart_refresh(self, data: data.FileData, extras: dict[str, Any]) -> data.FileData:
        """Show details of the incomplete upload.

        Args:
            data: The FileData object containing the upload metadata.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def multipart_update(self, data: data.FileData, upload: Upload, part: int, extras: dict[str, Any]) -> data.FileData:
        """Add data to the incomplete upload.

        Args:
            data: The FileData object containing the upload metadata.
            upload: The Upload object containing the content.
            part: Position of the given part among other parts, starting with 0.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def multipart_complete(self, data: data.FileData, extras: dict[str, Any]) -> data.FileData:
        """Verify file integrity and finalize incomplete upload.

        Args:
            data: The FileData object containing the upload metadata.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError

    def multipart_remove(self, data: data.FileData, extras: dict[str, Any]) -> bool:
        """Interrupt and remove incomplete upload.

        Args:
            data: The FileData object containing the upload metadata.
            extras: Additional metadata for the upload.
        """
        raise NotImplementedError


class Manager(StorageService):
    """Service responsible for maintenance file operations.

    `Storage` internally calls methods of this service. For example,
    `Storage.remove(data, **kwargs)` results in `Manager.remove(data, kwargs)`.

    Example:
        ```python
        class MyManager(Manager):
            capabilities = Capability.REMOVE
            def remove(
                self, data: FileData|FileData, extras: dict[str, Any]
            ) -> bool:
                os.remove(data.location)
                return True
        ```
    """

    def remove(self, data: data.FileData, extras: dict[str, Any]) -> bool:
        """Remove file from the storage.

        Args:
            data: The FileData object representing the file to remove.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def exists(self, data: data.FileData, extras: dict[str, Any]) -> bool:
        """Check if file exists in the storage.

        Args:
            data: The FileData object representing the file to check.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def compose(
        self, location: types.Location, data: Iterable[data.FileData], extras: dict[str, Any]
    ) -> data.FileData:
        """Combine multiple file inside the storage into a new one.

        Args:
            location: The destination location for the composed file.
            data: An iterable of FileData objects representing the files to combine.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def append(self, data: data.FileData, upload: Upload, extras: dict[str, Any]) -> data.FileData:
        """Append content to existing file.

        Args:
            data: The FileData object representing the file to append to.
            upload: The Upload object containing the content to append.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def copy(self, location: types.Location, data: data.FileData, extras: dict[str, Any]) -> data.FileData:
        """Copy file inside the storage.

        Args:
            location: The destination location for the copied file.
            data: The FileData object representing the file to copy.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def move(self, location: types.Location, data: data.FileData, extras: dict[str, Any]) -> data.FileData:
        """Move file to a different location inside the storage.

        Args:
            location: The destination location for the moved file.
            data: The FileData object representing the file to move.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        """List all locations(filenames) in storage.

        Args:
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def filtered_scan(self, prefix: str, glob: str, extras: dict[str, Any]) -> Iterable[str]:
        """List all locations(filenames) in storage that match prefix and glob.

        Args:
            prefix: The prefix to filter locations.
            glob: The glob pattern to filter locations.
            extras: Additional metadata for the operation.
        """
        names = self.scan(extras)
        if glob:
            names = fnmatch.filter(names, glob)

        if prefix:
            names = filter(lambda n: n.startswith(prefix), names)

        yield from names

    def analyze(self, location: types.Location, extras: dict[str, Any]) -> data.FileData:
        """Return details about location.

        Args:
            location: The location of the file to analyze.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def size(self, location: types.Location, extras: dict[str, Any]) -> int:
        """Return size of the file in bytes.

        Args:
            location: The location of the file.
            extras: Additional metadata for the operation.
        """
        return self.analyze(location, extras).size

    def hash(self, location: types.Location, extras: dict[str, Any]) -> str:
        """Return hash of the file.

        Args:
            location: The location of the file.
            extras: Additional metadata for the operation.
        """
        return self.analyze(location, extras).hash

    def content_type(self, location: types.Location, extras: dict[str, Any]) -> str:
        """Return MIME type of the file.

        Args:
            location: The location of the file.
            extras: Additional metadata for the operation.
        """
        return self.analyze(location, extras).content_type

    def signed(
        self, action: types.SignedAction, duration: int, location: types.Location, extras: dict[str, Any]
    ) -> str:
        """Make an URL for signed action.

        Args:
            action: The action to sign (e.g., "upload", "download").
            duration: The duration for which the signed URL is valid.
            location: The location of the file to sign.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError


class Reader(StorageService):
    """Service responsible for reading data from the storage.

    `Storage` internally calls methods of this service. For example,
    `Storage.stream(data, **kwargs)` results in `Reader.stream(data, kwargs)`.

    Example:
        ```python
        class MyReader(Reader):
            capabilities = Capability.STREAM

            def stream(
                self, data: data.FileData, extras: dict[str, Any]
            ) -> Iterable[bytes]:
                return open(data.location, "rb")
        ```
    """

    def stream(self, data: data.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        """Return byte-stream of the file content.

        Args:
            data: The FileData object representing the file to stream.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def content(self, data: data.FileData, extras: dict[str, Any]) -> bytes:
        """Return file content as a single byte object.

        Args:
            data: The FileData object representing the file to read.
            extras: Additional metadata for the operation.
        """
        return b"".join(self.stream(data, extras))

    def range(self, data: data.FileData, start: int, end: int | None, extras: dict[str, Any]) -> Iterable[bytes]:
        """Return slice of the file content.

        Args:
            data: The FileData object representing the file to read.
            start: The starting byte offset.
            end: The ending byte offset (inclusive).
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def permanent_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return permanent download link.

        Args:
            data: The FileData object representing the file.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def temporal_link(self, data: data.FileData, duration: int, extras: dict[str, Any]) -> str:
        """Return temporal download link.

        Args:
            data: The FileData object representing the file.
            duration: The duration for which the link is valid.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError

    def one_time_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return one-time download link.

        Args:
            data: The FileData object representing the file.
            extras: Additional metadata for the operation.
        """
        raise NotImplementedError


@dataclasses.dataclass()
class Settings:
    """Settings for the storage adapter."""

    type: str = ""
    """Type of the storage adapter."""

    name: str = "unknown"
    """Descriptive name of the storage used for debugging."""

    override_existing: bool = False
    """If file already exists, replace it with new content."""

    path: str = ""
    """Prefix for the file's location."""

    location_transformers: list[str] = cast("list[str]", dataclasses.field(default_factory=list))
    """List of transformations applied to the file location."""

    disabled_capabilities: list[str] = cast("list[str]", dataclasses.field(default_factory=list))
    """Capabilities that are not supported even if implemented."""

    initialize: bool = False
    """Prepare storage backend for uploads(create path, bucket, DB)"""

    skip_in_place_move: bool = True
    """Skip in-place move operations."""

    skip_in_place_copy: bool = True
    """Skip in-place copy operations."""

    _required_options: ClassVar[list[str]] = []
    _extra_settings: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))

    def __post_init__(self, **kwargs: Any):
        for attr in self._required_options:
            if not getattr(self, attr):
                raise exceptions.MissingStorageConfigurationError(self.name, attr)

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Settings:
        """Make settings object using dictionary as a source.

        Any unexpected options are extracted from the `data` to avoid
        initialization errors from dataclass constructor.

        Args:
            data: mapping with settings

        Returns:
            settings object built from data

        """
        # try:
        #     return cls(**settings)
        # except TypeError as err:
        #     raise exceptions.InvalidStorageConfigurationError(
        #         settings.get("name") or cls, str(err)
        #     ) from err

        sig = inspect.signature(cls)
        names = set(sig.parameters)

        valid = {}
        invalid = {}
        for k, v in data.items():
            if k in names:
                valid[k] = v
            else:
                invalid[k] = v

        valid.setdefault("_extra_settings", {}).update(invalid)
        cfg = cls(**valid)
        if invalid:
            log.warning(
                "Storage %s received unknown settings: %s",
                cfg.name,
                invalid,
            )
        return cfg


class Storage(ABC):  # noqa: B024
    """Base class for storage implementation.

    Args:
        settings: storage configuration

    Example:
        Extend base class to implement custom storage
        ```py
        class MyStorage(Storage):
            SettingsFactory = Settings
            UploaderFactory = Uploader
            ManagerFactory = Manager
            ReaderFactory = Reader
        ```
        then initialize it using required settings
        ```py
        my_storage = MyStorage({"option": "value"})
        ```
    """

    SettingsFactory: ClassVar[type[Settings]] = Settings
    """Factory class for storage settings."""
    UploaderFactory: ClassVar[type[Uploader]] = Uploader
    """Factory class for uploader service."""
    ManagerFactory: ClassVar[type[Manager]] = Manager
    """Factory class for manager service."""
    ReaderFactory: ClassVar[type[Reader]] = Reader
    """Factory class for reader service."""

    capabilities: Capability = Capability.NONE
    """Operations supported by storage. Computed from capabilities of
    services during storage initialization."""
    hidden: bool = False
    """Flag that marks unsafe/experimental storages."""

    @override
    def __str__(self) -> str:
        return self.settings.name

    @utils.ensure_setup
    def __init__(self, settings: Mapping[str, Any] | Settings, /):
        self.settings = self.configure(settings)
        self.uploader = self.make_uploader()
        self.manager = self.make_manager()
        self.reader = self.make_reader()

        self.capabilities = self.compute_capabilities()

    def make_uploader(self):
        """Initialize [uploader][file_keeper.Uploader] service."""
        return self.UploaderFactory(self)

    def make_manager(self):
        """Initialize [manager][file_keeper.Manager] service."""
        return self.ManagerFactory(self)

    def make_reader(self):
        """Initialize [reader][file_keeper.Reader] service."""
        return self.ReaderFactory(self)

    @classmethod
    def configure(cls, settings: Mapping[str, Any] | Settings) -> Settings:
        """Initialize storage configuration.

        This method is responsible for transforming mapping with options into
        storage's settings. It also can initialize additional services and
        perform extra work, like verifying that storage is ready to accept
        uploads.

        Args:
            settings: mapping with storage configuration

        Returns:
            initialized settings object
        """
        if isinstance(settings, Settings):
            return settings

        return cls.SettingsFactory.from_dict(settings)

    def compute_capabilities(self) -> utils.Capability:
        """Computes the capabilities of the storage based on its services.

        Combines the capabilities of the uploader, manager, and reader services,
        then excludes any capabilities that are listed in the storage settings as disabled.

        Returns:
            The combined capabilities of the storage.
        """
        result = self.uploader.capabilities | self.manager.capabilities | self.reader.capabilities

        for name in self.settings.disabled_capabilities:
            cap = name if isinstance(name, Capability) else Capability[name]
            result = result.exclude(cap)

        return result

    def supports(self, operation: utils.Capability) -> bool:
        """Check whether the storage supports operation.

        Args:
            operation: capability to check

        Returns:
            True if operation is supported, False otherwise
        """
        return self.capabilities.can(operation)

    def supports_synthetic(self, operation: utils.Capability, dest: Storage) -> bool:
        """Check if the storage can emulate operation using other operations.

        This method checks if the storage can perform the specified operation
        using a combination of other supported operations, often in conjunction
        with a destination storage.

        Synthetic operations are not stable and may change in future. They are
        not considered when computing capabilities of the storage. There are
        two main reasons to use them:

        * required operation involves two storage. For example, moving or
          copying file from one storage to another.
        * operation is not natively supported by the storage, but can be
          emulated using other operations. For example,
          [RANGE][file_keeper.Capability.RANGE] capability means that storage
          can return specified slice of the file. When this capability is not
          natively supported, storage can use
          [STREAM][file_keeper.Capability.STREAM] capability to read the whole
          file, returning only specified fragment. This is not efficient but
          still can solve certain problems.

        Args:
            operation: capability to check
            dest: destination storage for operations that require it

        Returns:
            True if operation is supported, False otherwise

        """
        if operation is Capability.RANGE:
            return self.supports(Capability.STREAM)

        if operation is Capability.COPY:
            return self.supports(Capability.STREAM) and dest.supports(
                Capability.CREATE,
            )

        if operation is Capability.MOVE:
            return self.supports(
                Capability.STREAM | Capability.REMOVE,
            ) and dest.supports(Capability.CREATE)

        if operation is Capability.COMPOSE:
            return self.supports(Capability.STREAM) and dest.supports(
                Capability.CREATE | Capability.APPEND | Capability.REMOVE
            )

        return False

    def full_path(self, location: types.Location, /, **kwargs: Any) -> str:
        """Compute path to the file from the storage's root.

        This method works as a shortcut for enabling `path` option. Whenever
        your custom storage works with location provided by user, wrap this
        location into this method to get full path:

        ```py
        class MyCustomReader:
            def stream(self, data: FileData, extras):
                real_location = self.storage.full_path(data_location)
                ...
        ```

        Args:
            location: location of the file object
            **kwargs: extra parameters for custom storages

        Returns:
            full path required to access location

        Raises:
            exceptions.LocationError: when location is outside of the storage's path
        """
        # Check for null bytes which could be used for injection
        if "\x00" in location:
            raise exceptions.LocationError(self, location)

        result = os.path.normpath(os.path.join(self.settings.path, location))
        if not result.startswith(self.settings.path):
            raise exceptions.LocationError(self, location)

        return result

    def prepare_location(self, location: str, sample: Upload | None = None, /, **kwargs: Any) -> types.Location:
        """Transform and sanitize location using configured functions.

        This method applies all transformations configured in
        [location_transformers][file_keeper.Settings.location_transformers]
        setting to the provided location. Each transformer is called in the
        order they are listed in the setting. The output of the previous
        transformer is passed as an input to the next one.

        Example:
            ```py
            location = storage.prepare_location(untrusted_location)
            ```

        Args:
            location: initial location provided by user
            sample: optional Upload object that can be used by transformers.
            **kwargs: additional parameters for transformers

        Returns:
            transformed location

        Raises:
            exceptions.LocationTransformerError: when transformer is not found

        """
        for name in self.settings.location_transformers:
            if transformer := location_transformers.get(name):
                location = transformer(location, sample, kwargs)

            else:
                raise exceptions.LocationTransformerError(name)

        return types.Location(location)

    def file_as_upload(self, data: data.FileData, **kwargs: Any) -> Upload:
        """Make an [Upload][file_keeper.Upload] with file content.

        Args:
            data: The FileData object to wrap into Upload
            **kwargs: Additional metadata for the upload.

        Returns:
            Upload object with file content
        """
        stream = self.stream(data, **kwargs)
        stream = cast(types.PStream, stream) if hasattr(stream, "read") else utils.IterableBytesReader(stream)

        return Upload(
            stream,
            data.location,
            data.size,
            data.content_type,
        )

    @requires_capability(Capability.CREATE)
    def upload(self, location: types.Location, upload: Upload, /, **kwargs: Any) -> data.FileData:
        """Upload file using single stream.

        Requires [CREATE][file_keeper.Capability.CREATE] capability.

        This is the simplest way to upload file into the storage. It uses
        single stream to transfer the whole file. If upload fails, no file is
        created in the storage.

        Content is not modified during upload, it is written as-is. And content
        can be of size 0, which results in empty file in the storage.

        When file already exists, behavior depends on
        [override_existing][file_keeper.Settings.override_existing] setting. If
        it is False, [ExistingFileError][file_keeper.exc.ExistingFileError] is
        raised. If it is True, existing file is replaced with new content. In
        this case, it is possible to lose existing file if upload fails. When
        adapter does not support removal, but supports overrides, this can be
        used to wipe the content of the file.

        Location can contain nested path as long as it does not go outside of
        the [path][file_keeper.Settings.path] from settings. For example, if
        `path` is set to `/data`, location can be `file.txt` or `2024/file.txt`
        but not `../file.txt` or `/etc/passwd`. Violating this rule leads to
        [LocationError][file_keeper.exc.LocationError].

        Args:
            location: The destination location for the upload.
            upload: The Upload object containing the file data.
            **kwargs: Additional metadata for the upload.

        Returns:
            FileData object with details about the uploaded file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                CREATE operation
            exceptions.ExistingFileError: when file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False
            exceptions.LocationError: when location is outside of the storage's path
        """
        return self.uploader.upload(location, upload, kwargs)

    @requires_capability(Capability.RESUMABLE)
    def resumable_start(self, location: types.Location, size: int, /, **kwargs: Any) -> data.FileData:
        """Prepare everything for resumable upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [RESUMABLE][file_keeper.Capability.RESUMABLE] capability.

        `content_type` and `hash` are optional. When any of those provided, it
        will be used to verify the integrity of the upload. If they are
        missing, the upload will be accepted without verification.

        Args:
            location: The destination location for the upload.
            size: The total size of the upload in bytes.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                RESUMABLE operation
        """
        return self.uploader.resumable_start(location, size, kwargs)

    @requires_capability(Capability.RESUMABLE)
    def resumable_refresh(self, data: data.FileData, /, **kwargs: Any) -> data.FileData:
        """Show details of the incomplete resumable upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [RESUMABLE][file_keeper.Capability.RESUMABLE] capability.

        Args:
            data: The FileData object containing the upload metadata.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                RESUMABLE operation
        """
        return self.uploader.resumable_refresh(data, kwargs)

    @requires_capability(Capability.RESUMABLE)
    def resumable_resume(self, data: data.FileData, upload: Upload, /, **kwargs: Any) -> data.FileData:
        """Resume the interrupted resumable upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [RESUMABLE][file_keeper.Capability.RESUMABLE] capability.

        Args:
            data: The FileData object containing the upload metadata.
            upload: The Upload object containing the content.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                RESUMABLE operation
        """
        return self.uploader.resumable_resume(data, upload, kwargs)

    @requires_capability(Capability.RESUMABLE)
    def resumable_remove(self, data: data.FileData, /, **kwargs: Any) -> bool:
        """Remove incomplete resumable upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [RESUMABLE][file_keeper.Capability.RESUMABLE] capability.

        Args:
            data: The FileData object containing the upload metadata.
            **kwargs: Additional metadata for the upload.

        Returns:
            True if upload was removed, False otherwise.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                RESUMABLE operation
        """
        return self.uploader.resumable_remove(data, kwargs)

    @requires_capability(Capability.MULTIPART)
    def multipart_start(self, location: types.Location, size: int, /, **kwargs: Any) -> data.FileData:
        """Prepare everything for multipart upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [MULTIPART][file_keeper.Capability.MULTIPART] capability.

        `content_type` and `hash` are optional. When any of those provided, it
        will be used to verify the integrity of the upload. If they are
        missing, the upload will be accepted without verification.

        Args:
            location: The destination location for the upload.
            size: The total size of the upload in bytes.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MULTIPART operation

        """
        return self.uploader.multipart_start(location, size, kwargs)

    @requires_capability(Capability.MULTIPART)
    def multipart_refresh(self, data: data.FileData, /, **kwargs: Any) -> data.FileData:
        """Show details of the incomplete upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [MULTIPART][file_keeper.Capability.MULTIPART] capability.

        Args:
            data: The FileData object containing the upload metadata.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MULTIPART operation

        """
        return self.uploader.multipart_refresh(data, kwargs)

    @requires_capability(Capability.MULTIPART)
    def multipart_update(self, data: data.FileData, upload: Upload, part: int, /, **kwargs: Any) -> data.FileData:
        """Add data to the incomplete upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [MULTIPART][file_keeper.Capability.MULTIPART] capability.

        Args:
            data: The FileData object containing the upload metadata.
            upload: The Upload object containing the content.
            part: Position of the given part among other parts, starting with 0.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MULTIPART operation

        """
        return self.uploader.multipart_update(data, upload, part, kwargs)

    @requires_capability(Capability.MULTIPART)
    def multipart_complete(self, data: data.FileData, /, **kwargs: Any) -> data.FileData:
        """Verify file integrity and finalize incomplete upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [MULTIPART][file_keeper.Capability.MULTIPART] capability.

        Args:
            data: The FileData object containing the upload metadata.
            **kwargs: Additional metadata for the upload.

        Returns:
            details about the upload.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MULTIPART operation

        """
        return self.uploader.multipart_complete(data, kwargs)

    @requires_capability(Capability.MULTIPART)
    def multipart_remove(self, data: data.FileData, /, **kwargs: Any) -> bool:
        """Interrupt and remove incomplete upload.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [MULTIPART][file_keeper.Capability.MULTIPART] capability.

        Args:
            data: The FileData object containing the upload metadata.
            **kwargs: Additional metadata for the upload.

        Returns:
            True if upload was removed, False otherwise.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MULTIPART operation

        """
        return self.uploader.multipart_remove(data, kwargs)

    @requires_capability(Capability.EXISTS)
    def exists(self, data: data.FileData, /, **kwargs: Any) -> bool:
        """Check if file exists in the storage.

        Requires [EXISTS][file_keeper.Capability.EXISTS] capability.

        Args:
            data: The FileData object representing the file to check.
            **kwargs: Additional metadata for the operation.

        Returns:
            True if file exists, False otherwise.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                EXISTS operation
        """
        return self.manager.exists(data, kwargs)

    @requires_capability(Capability.REMOVE)
    def remove(self, data: data.FileData, /, **kwargs: Any) -> bool:
        """Remove file from the storage.

        Requires [REMOVE][file_keeper.Capability.REMOVE] capability.

        Args:
            data: The FileData object representing the file to remove.
            **kwargs: Additional metadata for the operation.

        Returns:
            True if file was removed, False otherwise.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                REMOVE operation
        """
        return self.manager.remove(data, kwargs)

    @requires_capability(Capability.SCAN)
    def scan(self, **kwargs: Any) -> Iterable[str]:
        """List all locations(filenames) in storage.

        Requires [SCAN][file_keeper.Capability.SCAN] capability.

        This operation lists all locations (filenames) in the storage if they
        start with the configured [path][file_keeper.Settings.path]. If path is
        empty, all locations are listed. Locations that match path
        partially(e.g. location `nested_dir` overlaps with path `nested`) are
        not listed.

        Args:
            **kwargs: Additional metadata for the operation.

        Returns:
            An iterable of location strings.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                SCAN operation
        """
        return self.manager.scan(kwargs)

    @requires_capability(Capability.SCAN)
    def filtered_scan(self, /, prefix: str = "", glob: str = "", **kwargs: Any) -> Iterable[str]:
        """List all locations(filenames) in storage that match prefix and glob.

        Requires [SCAN][file_keeper.Capability.SCAN] capability.

        This operation lists all locations (filenames) in the storage if they
        start with the specified prefix and match the provided glob pattern. If
        no prefix is provided, it defaults to an empty string, meaning all
        locations are considered. If no glob pattern is provided, it defaults to
        an empty string, which matches all filenames.

        Glob parameter supports standard Unix shell-style wildcards:

        * `*` matches everything
        * `?` matches any single character
        * `[seq]` matches any character in seq
        * `[!seq]` matches any character not in seq

        Adapters may choose to support different set of wildcards. Check
        adapter specific documentation for details.

        Args:
            prefix: The prefix to filter locations.
            glob: The glob pattern to filter locations.
            **kwargs: Additional metadata for the operation.

        Returns:
            An iterable of location strings.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                SCAN operation
        """
        return self.manager.filtered_scan(prefix, glob, kwargs)

    @requires_capability(Capability.ANALYZE)
    def analyze(self, location: types.Location, /, **kwargs: Any) -> data.FileData:
        """Return file details.

        Requires [ANALYZE][file_keeper.Capability.ANALYZE] capability.

        [FileData][file_keeper.FileData] produced by this operation is the same
        as data produced by the [upload()][file_keeper.Storage.upload].

        Attempt to analyze non-existing location leads to
        [MissingFileError][file_keeper.exc.MissingFileError]

        Args:
            location: The location of the file to analyze.
            **kwargs: Additional metadata for the operation.

        Returns:
            details about the file.

        Raises:
            exceptions.MissingFileError: when location does not exist
            exceptions.UnsupportedOperationError: when storage does not support
                ANALYZE operation
        """
        return self.manager.analyze(location, kwargs)

    @requires_capability(Capability.ANALYZE)
    def size(self, location: types.Location, /, **kwargs: Any) -> int:
        """Return the size of the file.

        Requires [ANALYZE][file_keeper.Capability.ANALYZE] capability.

        Args:
            location: The location of the file to analyze.
            **kwargs: Additional metadata for the operation.

        Returns:
            size of the file

        Raises:
            exceptions.MissingFileError: when location does not exist
            exceptions.UnsupportedOperationError: when storage does not support
                ANALYZE operation
        """
        return self.manager.size(location, kwargs)

    @requires_capability(Capability.ANALYZE)
    def hash(self, location: types.Location, /, **kwargs: Any) -> str:
        """Return the hash of the file.

        Requires [ANALYZE][file_keeper.Capability.ANALYZE] capability.

        Args:
            location: The location of the file to analyze.
            **kwargs: Additional metadata for the operation.

        Returns:
            hash of the file

        Raises:
            exceptions.MissingFileError: when location does not exist
            exceptions.UnsupportedOperationError: when storage does not support
                ANALYZE operation
        """
        return self.manager.hash(location, kwargs)

    @requires_capability(Capability.ANALYZE)
    def content_type(self, location: types.Location, /, **kwargs: Any) -> str:
        """Return the MIME type of the file.

        Requires [ANALYZE][file_keeper.Capability.ANALYZE] capability.

        Args:
            location: The location of the file to analyze.
            **kwargs: Additional metadata for the operation.

        Returns:
            MIME type of the file

        Raises:
            exceptions.MissingFileError: when location does not exist
            exceptions.UnsupportedOperationError: when storage does not support
                ANALYZE operation
        """
        return self.manager.content_type(location, kwargs)

    @requires_capability(Capability.SIGNED)
    def signed(
        self,
        action: types.SignedAction,
        duration: int,
        location: types.Location,
        **kwargs: Any,
    ) -> str:
        """Make an URL for signed action.

        /// warning
        This operation is not stabilized yet.
        ///

        Requires [SIGNED][file_keeper.Capability.SIGNED] capability.

        This operation creates a signed URL that allows performing the
        specified action (e.g., "upload", "download") on the given location for
        a limited duration. The signed URL is typically used to grant temporary
        access to a file without requiring authentication. The URL includes a
        signature that verifies its authenticity and validity.

        Depending on the action, user is expected to use the URL in
        different ways:

        * upload - use HTTP PUT request to send the file content to the URL.
        * download - use HTTP GET request to retrieve the file content from
            the URL.
        * delete - use HTTP DELETE request to remove the file at the URL.

        Check adapter specific implementation for details. Some actions may not
        be supported or require additional information. For example, upload
        with Azure Blob Storage requires header `x-ms-blob-type: BlockBlob`.

        Args:
            action: The action to sign (upload, download, delete).
            duration: The duration for which the signed URL is valid.
            location: The location of the file to sign.
            **kwargs: Additional metadata for the operation.

        Returns:
            A signed URL as a string.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                SIGNED operation
        """
        return self.manager.signed(action, duration, location, kwargs)

    @requires_capability(Capability.STREAM)
    def stream(self, data: data.FileData, /, **kwargs: Any) -> Iterable[bytes]:
        """Return byte-stream of the file content.

        Requires [STREAM][file_keeper.Capability.STREAM] capability.

        Returns iterable that yields chunks of bytes from the file. The size of
        each chunk depends on the storage implementation. The iterable can be
        used in a for loop or converted to a list to get all chunks at once.

        Args:
            data: The FileData object representing the file to stream.
            **kwargs: Additional metadata for the operation.

        Returns:
            An iterable yielding chunks of bytes from the file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                STREAM operation
            exceptions.MissingFileError: when file does not exist

        """
        return self.reader.stream(data, kwargs)

    @requires_capability(Capability.RANGE)
    def range(self, data: data.FileData, start: int = 0, end: int | None = None, /, **kwargs: Any) -> Iterable[bytes]:
        """Return slice of the file content.

        Requires [RANGE][file_keeper.Capability.RANGE] capability.

        This operation slices the file content from the specified start byte
        offset to the end byte offset (exclusive). If `end` is None, the slice
        extends to the end of the file. The operation returns an iterable that
        yields chunks of bytes from the specified range. The size of each chunk
        depends on the storage implementation. The iterable can be used in a for
        loop or converted to a list to get all chunks at once.

        Unlike python slices, this operation does not expect negative indexes,
        but certain adapters may support it.

        Attempt to get a range from non-existing file leads to
        [MissingFileError][file_keeper.exc.MissingFileError].

        Args:
            data: The FileData object representing the file to read.
            start: The starting byte offset.
            end: The ending byte offset (inclusive).
            **kwargs: Additional metadata for the operation.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                RANGE operation
            exceptions.MissingFileError: when file does not exist

        """
        return self.reader.range(data, start, end, kwargs)

    def range_synthetic(
        self, data: data.FileData, start: int = 0, end: int | None = None, /, **kwargs: Any
    ) -> Iterable[bytes]:
        """Generic implementation of range operation that relies on [STREAM][file_keeper.Capability.STREAM].

        This method provides a generic implementation of the range operation
        using the STREAM capability. It reads the file in chunks and yields only
        the specified range of bytes.

        Args:
            data: The FileData object representing the file to read.
            start: The starting byte offset.
            end: The ending byte offset (inclusive).
            **kwargs: Additional metadata for the operation.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                STREAM operation
            exceptions.MissingFileError: when file does not exist
        """
        if end is None:
            end = data.size + 1

        end -= start
        if end <= 0:
            return

        for chunk in self.stream(data, **kwargs):
            if start > 0:
                start -= len(chunk)
                if start < 0:
                    chunk = chunk[start:]  # noqa: PLW2901
                else:
                    continue

            yield chunk[:end]
            end -= len(chunk)
            if end <= 0:
                break

    @requires_capability(Capability.STREAM)
    def content(self, data: data.FileData, /, **kwargs: Any) -> bytes:
        """Return file content as a single byte object.

        Requires [STREAM][file_keeper.Capability.STREAM] capability.

        Returns complete file content as a single bytes object. This method
        internally uses [stream()][file_keeper.Storage.stream] method to read
        the file content in chunks and then combines those chunks into a single
        bytes object.

        Args:
            data: The FileData object representing the file to read.
            **kwargs: Additional metadata for the operation.

        Returns:
            The complete file content as a single bytes object.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                STREAM operation
            exceptions.MissingFileError: when file does not exist
        """
        return self.reader.content(data, kwargs)

    @requires_capability(Capability.APPEND)
    def append(self, data: data.FileData, upload: Upload, /, **kwargs: Any) -> data.FileData:
        """Append content to existing file.

        Requires [APPEND][file_keeper.Capability.APPEND] capability.

        Args:
            data: The FileData object representing the file to append to.
            upload: The Upload object containing the content to append.
            **kwargs: Additional metadata for the operation.

        Returns:
            Updated FileData object with details about the file after appending.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                APPEND operation
            exceptions.MissingFileError: when file does not exist

        """
        return self.manager.append(data, upload, kwargs)

    @requires_capability(Capability.COPY)
    def copy(self, location: types.Location, data: data.FileData, /, **kwargs: Any) -> data.FileData:
        """Copy file inside the storage.

        Requires [COPY][file_keeper.Capability.COPY] capability.

        This operation creates a duplicate of the specified file at a new
        location within the same storage. The copied file retains the same
        content and metadata as the original file.

        If file already exists at the destination location, behavior depends on
        [override_existing][file_keeper.Settings.override_existing] setting. If
        it is False, [ExistingFileError][file_keeper.exc.ExistingFileError] is
        raised. If it is True, existing file is replaced with the copied file. In
        this case, it is possible to lose existing file if copy fails.

        Args:
            location: The destination location for the copied file.
            data: The FileData object representing the file to copy.
            **kwargs: Additional metadata for the operation.

        Returns:
            FileData object with details about the copied file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                COPY operation
            exceptions.MissingFileError: when source file does not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False
        """
        if location == data.location and self.settings.skip_in_place_copy:
            return data

        return self.manager.copy(location, data, kwargs)

    def copy_synthetic(
        self, location: types.Location, data: data.FileData, dest_storage: Storage, /, **kwargs: Any
    ) -> data.FileData:
        """Generic implementation of the copy operation that relies on [CREATE][file_keeper.Capability.CREATE].

        This method provides a generic implementation of the copy operation
        using the CREATE capability of the destination storage and the STREAM
        capability of the source storage. It reads the file content from the
        source storage and uploads it to the destination storage.

        Args:
            location: The destination location for the copied file.
            data: The FileData object representing the file to copy.
            dest_storage: The destination storage
            **kwargs: Additional metadata for the operation.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                CREATE operation
            exceptions.MissingFileError: when source file does not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False

        """
        return dest_storage.upload(
            location,
            self.file_as_upload(data, **kwargs),
            **kwargs,
        )

    @requires_capability(Capability.MOVE)
    def move(self, location: types.Location, data: data.FileData, /, **kwargs: Any) -> data.FileData:
        """Move file to a different location inside the storage.

        Requires [MOVE][file_keeper.Capability.MOVE] capability.

        This operation relocates the specified file to a new location within
        the same storage. After the move operation, the file will no longer
        exist at its original location.

        Args:
            location: The destination location for the moved file.
            data: The FileData object representing the file to move.
            **kwargs: Additional metadata for the operation.

        Returns:
            FileData object with details about the moved file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                MOVE operation
            exceptions.MissingFileError: when source file does not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False

        """
        if location == data.location and self.settings.skip_in_place_move:
            return data
        return self.manager.move(location, data, kwargs)

    def move_synthetic(
        self, location: types.Location, data: data.FileData, dest_storage: Storage, /, **kwargs: Any
    ) -> data.FileData:
        """Generic implementation of move operation.

        Relies on [CREATE][file_keeper.Capability.CREATE] and
        [REMOVE][file_keeper.Capability.REMOVE].

        Args:
            location: The destination location for the moved file.
            data: The FileData object representing the file to move.
            dest_storage: The destination storage
            **kwargs: Additional metadata for the operation.

        Returns:
            FileData object with details about the moved file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                CREATE or REMOVE operation
            exceptions.MissingFileError: when source file does not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False

        """
        result = dest_storage.upload(location, self.file_as_upload(data, **kwargs), **kwargs)
        self.remove(data)
        return result

    @requires_capability(Capability.COMPOSE)
    def compose(self, location: types.Location, /, *files: data.FileData, **kwargs: Any) -> data.FileData:
        """Combine multiple files into a new file.

        Requires [COMPOSE][file_keeper.Capability.COMPOSE] capability.

        This operation combines multiple source files into a single destination
        file. The content of the source files is concatenated in the order they
        are provided to form the content of the new file.

        Args:
            location: The destination location for the composed file.
            *files: FileData objects representing the files to combine.
            **kwargs: Additional metadata for the operation.

        Returns:
            FileData object with details about the composed file.

        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                COMPOSE operation
            exceptions.MissingFileError: when any of source files do not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False
        """
        return self.manager.compose(location, files, kwargs)

    def compose_synthetic(
        self, location: types.Location, dest_storage: Storage, /, *files: data.FileData, **kwargs: Any
    ) -> data.FileData:
        """Generic composition that relies on [APPEND][file_keeper.Capability.APPEND].

        This method provides a generic implementation of the compose operation
        using the APPEND capability of the destination storage. It creates an
        empty file at the destination location and then appends the content of
        each source file to it in sequence.

        Args:
            location: The destination location for the composed file.
            dest_storage: The destination storage
            *files: FileData objects representing the files to combine.
            **kwargs: Additional metadata for the operation.


        Raises:
            exceptions.UnsupportedOperationError: when storage does not support
                APPEND operation
            exceptions.MissingFileError: when any of source files do not exist
            exceptions.ExistingFileError: when destination file already exists and
                [override_existing][file_keeper.Settings.override_existing] is False

        """
        result = dest_storage.upload(location, make_upload(b""), **kwargs)

        # when first append succeeded with the fragment of the file added
        # in the storage, and the following append failed, this incomplete
        # fragment must be removed.
        #
        # Expected reasons of failure are:
        #
        # * one of the source files is missing
        # * file will go over the size limit after the following append
        try:
            for item in files:
                result = dest_storage.append(
                    result,
                    self.file_as_upload(item, **kwargs),
                    **kwargs,
                )
        except (exceptions.MissingFileError, exceptions.UploadError):
            self.remove(result, **kwargs)
            raise

        return result

    def one_time_link(self, data: data.FileData, /, **kwargs: Any) -> str | None:
        """Return one-time download link.

        Requires [LINK_ONE_TIME][file_keeper.Capability.LINK_ONE_TIME] capability.

        Args:
            data: The FileData object representing the file.
            **kwargs: Additional metadata for the operation.

        Returns:
            A one-time download link as a string, or None if not supported.
        """
        if self.supports(Capability.LINK_ONE_TIME):
            return self.reader.one_time_link(data, kwargs)

    def temporal_link(self, data: data.FileData, duration: int, /, **kwargs: Any) -> str | None:
        """Return temporal download link.

        Requires [LINK_TEMPORAL][file_keeper.Capability.LINK_TEMPORAL] capability.

        Args:
            data: The FileData object representing the file.
            duration: The duration for which the link is valid.
            **kwargs: Additional metadata for the operation.

        Returns:
            A temporal download link as a string, or None if not supported.
        """
        if self.supports(Capability.LINK_TEMPORAL):
            return self.reader.temporal_link(data, duration, kwargs)

    def permanent_link(self, data: data.FileData, /, **kwargs: Any) -> str | None:
        """Return permanent download link.

        Requires [LINK_PERMANENT][file_keeper.Capability.LINK_PERMANENT] capability.

        Args:
            data: The FileData object representing the file.
            **kwargs: Additional metadata for the operation.

        Returns:
            A permanent download link as a string, or None if not supported.
        """
        if self.supports(Capability.LINK_PERMANENT):
            return self.reader.permanent_link(data, kwargs)


@utils.ensure_setup
def make_storage(name: str, settings: dict[str, Any]) -> Storage:
    """Initialize storage instance with specified settings.

    Storage adapter is defined by `type` key of the settings. The rest of
    settings depends on the specific adapter.

    Example:
        ```py
        storage = make_storage("memo", {"type": "file_keeper:memory"})
        ```

    Args:
        name: name of the storage
        settings: configuration for the storage

    Returns:
        storage instance

    Raises:
        exceptions.UnknownAdapterError: storage adapter is not registered


    """
    adapter_type: Any = settings.get("type")
    adapter = adapters.get(adapter_type)
    if not adapter:
        raise exceptions.UnknownAdapterError(adapter_type)

    settings.setdefault("name", name)

    return adapter(settings)


@utils.ensure_setup
def list_adapters() -> list[str]:
    """List all registered storage adapters.

    Returns:
        list of registered storage adapter types
    """
    return list(adapters)


def get_storage(name: str, settings: dict[str, Any] | None = None) -> Storage:
    """Get storage from the pool.

    If storage accessed for the first time, it's initialized and added to the
    pool. After that the same storage is returned every time the function is
    called with the given name.

    Settings are required only for initialization, so you can omit them if you
    are sure that storage exists. Additionally, if `settings` are not specified
    but storage is missing from the pool, file-keeper makes an attempt to
    initialize storage using global configuration. Global configuration can be
    provided as:

    * `FILE_KEEPER_CONFIG` environment variable that points to a file with configuration
    * `.file-keeper.json` in the current directory hierarchy
    * `file-keeper/file-keeper.json` in the user's config directory(usually,
      `~/.config/`) when [platformdirs](https://pypi.org/project/platformdirs/)
      installed in the environment, for example via `pip install
      'file-keeper[user_config]'` extras.

    File must contain storage configuration provided in format

    ```json5
    {
        "storages": {
            "my_storage": {  # (1)!
                "type": "file_keeper:memory"  # (2)!
            }
        }
    }
    ```

    1. Name of the storage
    2. Options for the storage

    JSON configuration is used by default, because python has built-in JSON
    support. Additional file extensions are checked if environment contains
    corresponding package:

    | Package | Extension |
    |---|---|
    | [tomllib](https://docs.python.org/3/library/tomllib.html) | `.toml` |
    | [tomli](https://pypi.org/project/tomli/) | `.toml` |
    | [pyyaml](https://pypi.org/project/PyYAML/) | `.yaml`, `.yml` |

    Extensions are checked in order `.toml`, `.yaml`, `.yml`, `.json`.

    Example:
        If storage accessed for the first time, settings are required

        ```pycon
        >>> storage = get_storage("memory", {"type": "file_keeper:memory"})
        >>> storage
        <file_keeper.default.adapters.memory.MemoryStorage object at 0x...>

        ```

        and the same storage is returned every time in subsequent calls

        ```pycon
        >>> cached = get_storage("memory")
        >>> storage is cached
        True

        ```

        but if storage does not exist and settings are omitted, exception is raised

        ```pycon
        >>> get_storage("new-memory")
        Traceback (most recent call last):
            ...
        file_keeper.core.exceptions.UnknownStorageError: Storage new-memory is not configured

        ```

    Args:
        name: name of the storage
        settings: configuration for the storage

    Returns:
        storage instance

    Raises:
        exceptions.UnknownStorageError: storage with the given name is not configured
    """
    if name not in storages:
        if settings is None:
            config_file = os.getenv("FILE_KEEPER_CONFIG")

            if not config_file and (config_dir := user_config_dir("file-keeper")):
                config_file = _get_config_file_name(config_dir)

            if not config_file:
                path = pathlib.Path().absolute()

                while len(path.parts) > 1:
                    if config_file := _get_config_file_name(str(path), hidden=True):
                        break

                    path = path.parent

            if config_file:
                log.debug("Load configuration from %s", config_file)
                cfg = _load_config_file(config_file)
                settings = cfg.get("storages", {}).get(name)

            if not settings:
                raise exceptions.UnknownStorageError(name)

        storages.register(name, make_storage(name, settings))

    return storages[name]


def _get_config_file_name(path: str, /, hidden: bool = False):
    basename = ".file-keeper" if hidden else "file-keeper"
    plain_name = os.path.join(path, basename)

    if toml and os.path.exists(f"{plain_name}.toml"):
        return f"{plain_name}.toml"

    if yaml:
        if os.path.exists(f"{plain_name}.yaml"):
            return f"{plain_name}.yaml"

        if os.path.exists(f"{plain_name}.yml"):
            return f"{plain_name}.yml"

    if os.path.exists(f"{plain_name}.json"):
        return f"{plain_name}.json"


def _load_config_file(filename: str) -> dict[str, Any]:
    ext = os.path.splitext(filename)[1].lower()

    if ext == ".toml":
        with open(filename, "rb") as src:
            return toml.load(src)  # pyright: ignore[reportUnknownVariableType, reportOptionalMemberAccess]

    if ext in [".yaml", ".yml"]:
        with open(filename, "rb") as src:
            return yaml.full_load(src)  # pyright: ignore[reportOptionalMemberAccess]

    if ext == ".json":
        with open(filename, "rb") as src:
            return json.load(src)

    raise TypeError(filename)



"""Memory adapter."""

from __future__ import annotations

import dataclasses
import logging
import os
from collections.abc import Iterable, MutableMapping
from io import BytesIO
from typing import Any, cast

import magic
from typing_extensions import override

import file_keeper as fk

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for memory storage."""

    bucket: MutableMapping[str, bytes] = cast("dict[str, bytes]", dataclasses.field(default_factory=dict))
    """Container for uploaded objects."""


class Uploader(fk.Uploader):
    """Memory uploader."""

    storage: MemoryStorage
    capabilities: fk.Capability = fk.Capability.CREATE | fk.Capability.RESUMABLE

    @override
    def upload(self, location: fk.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        reader = upload.hashing_reader()
        if filepath in self.storage.settings.bucket and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        self.storage.settings.bucket[filepath] = reader.read()

        return fk.FileData(location, upload.size, upload.content_type, hash=reader.get_hash())

    @override
    def resumable_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        self.storage.settings.bucket[filepath] = b""

        return fk.FileData.from_dict(
            extras, size=size, location=location, storage_data={"resumable": True, "memory": {"uploaded": 0}}
        )

    @override
    def resumable_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)
        bucket = self.storage.settings.bucket

        if filepath not in bucket:
            raise fk.exc.MissingFileError(self.storage, data.location)

        result = fk.FileData.from_object(data)
        result.storage_data["memory"]["uploaded"] = len(bucket[filepath])

        if result.size == result.storage_data["memory"]["uploaded"]:
            reader = fk.HashingReader(BytesIO(bucket[filepath]))
            reader.exhaust()

            hash = reader.get_hash()
            result = fk.FileData.from_object(result, hash=hash)

            result.storage_data.pop("memory")
            result.storage_data.pop("resumable")

        return result

    @override
    def resumable_resume(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        bucket = self.storage.settings.bucket

        if filepath not in bucket:
            raise fk.exc.MissingFileError(self.storage, data.location)

        expected_size = upload.size + len(bucket[filepath])
        if expected_size > data.size:
            raise fk.exc.UploadOutOfBoundError(expected_size, data.size)

        result = fk.FileData.from_object(data)

        bucket[filepath] += upload.stream.read()
        result.storage_data["memory"]["uploaded"] = expected_size

        size = len(bucket[filepath])
        if result.size == size:
            reader = fk.HashingReader(BytesIO(bucket[filepath]))
            reader.exhaust()

            hash = reader.get_hash()
            result = fk.FileData.from_object(result, hash=hash)

            result.storage_data.pop("memory")
            result.storage_data.pop("resumable")

        return result

    @override
    def resumable_remove(self, data: fk.FileData, extras: dict[str, Any]):
        return self.storage.remove(data, **extras)


class Manager(fk.Manager):
    """Memory manager."""

    storage: MemoryStorage
    capabilities: fk.Capability = (
        fk.Capability.ANALYZE
        | fk.Capability.SCAN
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.APPEND
        | fk.Capability.COMPOSE
        | fk.Capability.EXISTS
        | fk.Capability.REMOVE
    )

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)

        bucket = self.storage.settings.bucket
        result = bucket.pop(filepath, None)
        return result is not None

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)

        bucket = self.storage.settings.bucket
        return filepath in bucket

    @override
    def compose(self, location: fk.Location, data: Iterable[fk.FileData], extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)

        bucket = self.storage.settings.bucket
        if filepath in bucket and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        result = b""
        for item in data:
            part_path = self.storage.full_path(item.location)
            if part_path not in bucket:
                raise fk.exc.MissingFileError(self.storage, item.location)
            result += bucket[part_path]

        bucket[filepath] = result

        return self.analyze(location, extras)

    @override
    def append(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        bucket = self.storage.settings.bucket
        if filepath not in bucket:
            raise fk.exc.MissingFileError(self.storage, data.location)

        bucket[filepath] += upload.stream.read()
        return self.analyze(data.location, extras)

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)

        bucket = self.storage.settings.bucket
        if filepath in bucket and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        sourcepath = self.storage.full_path(data.location)
        if sourcepath not in bucket:
            raise fk.exc.MissingFileError(self.storage, location)

        bucket[filepath] = bucket[sourcepath]
        return self.analyze(location, extras)

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        result = self.copy(location, data, extras)
        self.remove(data, extras)
        return result

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        # create a copy of keys to avoid "dictionary changed size during
        # iteration" error
        keys = list(self.storage.settings.bucket)
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"
        for key in keys:
            if key.startswith(path):
                yield os.path.relpath(key, path)

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)

        try:
            content = self.storage.settings.bucket[filepath]
        except KeyError as err:
            raise fk.exc.MissingFileError(self.storage, location) from err

        reader = fk.HashingReader(BytesIO(content))
        content_type = magic.from_buffer(next(reader, b""), True)

        return fk.FileData(location, len(content), content_type, reader.get_hash())


class Reader(fk.Reader):
    """Memory reader."""

    storage: MemoryStorage
    capabilities: fk.Capability = fk.Capability.READER_CAPABILITIES

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        filepath = self.storage.full_path(data.location)

        try:
            return [self.storage.settings.bucket[filepath]]
        except KeyError as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err

    @override
    def range(self, data: fk.FileData, start: int, end: int | None, extras: dict[str, Any]) -> Iterable[bytes]:
        filepath = self.storage.full_path(data.location)
        try:
            return [self.storage.settings.bucket[filepath][start:end]]
        except KeyError as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err


class MemoryStorage(fk.Storage):
    """Storage files in-memory.

    This storage adapter keeps files in memory using a dictionary. It is
    mainly useful for testing purposes. It is not recommended to use it in
    production as all files will be lost when the application stops.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:memory",
        "override_existing": False,
    }
    storage = fk.make_storage("memory", settings)
    ```

    Note:
    * The `override_existing` setting controls whether existing files can be
      overridden. If set to `False`, attempting to upload a file to a location
      that already exists will raise an `ExistingFileError`.
    * The `bucket` setting is a dictionary that holds the uploaded files in
        memory. It is automatically initialized as an empty dictionary if not
        provided.
    * This storage adapter does not support persistence. All files are lost
      when the application stops.
    """

    settings: Settings

    SettingsFactory = Settings
    UploaderFactory = Uploader
    ReaderFactory = Reader
    ManagerFactory = Manager



"""Obstore adapter."""

from __future__ import annotations

import dataclasses
import os
from collections.abc import Iterable
from typing import Any, cast

import magic
import obstore
from typing_extensions import override

import file_keeper as fk


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Obstore storage settings."""

    params: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))
    """Parameters for obstore store initialization."""
    url: str = ""
    """URL of obstore store."""
    store: obstore.store.ObjectStore = None  # pyright: ignore[reportAssignmentType]
    """Existing obstore store."""

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        self.path = self.path.strip("/")

        if not self.store:
            if not self.url:
                raise fk.exc.MissingStorageConfigurationError(self.name, "url")

            try:
                self.store = obstore.store.from_url(self.url, **self.params)
            except ValueError as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    str(err),
                ) from err


class Uploader(fk.Uploader):
    """FsSpec uploader."""

    storage: ObjectStoreStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        store = self.storage.settings.store
        filepath = self.storage.full_path(location)

        if not self.storage.settings.override_existing:
            try:
                store.head(filepath)
            except FileNotFoundError:
                pass
            else:
                raise fk.exc.ExistingFileError(self.storage, location)

        with obstore.open_writer(store, filepath) as fobj:
            reader = upload.hashing_reader()
            for chunk in reader:
                fobj.write(chunk)

        return fk.FileData(
            location=location,
            size=reader.position,
            content_type=upload.content_type,
            hash=reader.get_hash(),
        )


class Reader(fk.Reader):
    """FsSpec reader."""

    storage: ObjectStoreStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        store = self.storage.settings.store
        filepath = self.storage.full_path(data.location)

        try:
            resp = store.get(filepath)
            return resp.stream()

        except FileNotFoundError as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err


class Manager(fk.Manager):
    """FsSpec manager."""

    storage: ObjectStoreStorage
    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.ANALYZE
    )

    @override
    def copy(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        store = self.storage.settings.store

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        store.copy(src_location, dest_location, overwrite=self.storage.settings.override_existing)

        return fk.FileData.from_object(data, location=location)

    @override
    def move(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        store = self.storage.settings.store

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        store.rename(src_location, dest_location, overwrite=self.storage.settings.override_existing)

        return fk.FileData.from_object(data, location=location)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        store = self.storage.settings.store
        try:
            store.head(filepath)
        except FileNotFoundError:
            return False

        return True

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        try:
            src = self.storage.stream(fk.FileData(location), **extras)
        except FileNotFoundError as err:
            raise fk.exc.MissingFileError(self.storage, location) from err

        reader = fk.HashingReader(fk.IterableBytesReader(src))
        content_type = magic.from_buffer(next(reader, b""), True)
        reader.exhaust()

        return fk.FileData(location, size=reader.position, content_type=content_type, hash=reader.get_hash())

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        store = self.storage.settings.store

        if not self.exists(data, extras):
            return False

        store.delete(filepath)
        return True

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.full_path(fk.Location(""))

        store = self.storage.settings.store

        for info in store.list(path):  # pyright: ignore[reportUnknownVariableType]
            for name in self._scan_unwrap(info):
                yield os.path.relpath(name, path)

    def _scan_unwrap(self, info: Any) -> Iterable[str]:
        if isinstance(info, str):
            yield info
        elif isinstance(info, dict):
            yield info["path"]

        elif isinstance(info, list):
            for item in info:  # pyright: ignore[reportUnknownVariableType]
                yield from self._scan_unwrap(item)


class ObjectStoreStorage(fk.Storage):
    """Obstore storage adapter.

    This storage uses the [obstore](https://pypi.org/project/obstore/) library to provide
    access to various object storage backends. It supports any backend supported by
    obstore, including local filesystem, S3, Google Cloud Storage, Azure Blob Storage,
    and more.

    To use this storage, you need to provide the `url` setting in the configuration,
    which specifies the URL of the object store. The URL should be in the format
    supported by obstore. For example, to use a local filesystem as the
    object store, you can use the `file:///path/to/directory` URL.


    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:obstore",
        "url": "file:///tmp/file-keeper",
    }

    storage = fk.make_storage("obstore", settings)
    ```

    Note:
    * The `url` setting is required unless an existing `store` is provided.
    * The `params` setting is optional and can be used to provide additional parameters
      to the `obstore.store.from_url` function.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    @override
    def compute_capabilities(self) -> fk.Capability:
        return super().compute_capabilities()



"""AWS S3 adapter."""

from __future__ import annotations

import dataclasses
import os
import re
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, ClassVar
from urllib.parse import urlparse, urlunparse

import boto3
from typing_extensions import override

import file_keeper as fk

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3.type_defs import CreateMultipartUploadRequestTypeDef

RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")
HTTP_RESUME = 308


@dataclasses.dataclass()
class Settings(fk.Settings):
    """AWS S3 settings."""

    bucket: str = ""
    """Name of the storage bucket."""
    client: S3Client = None  # pyright: ignore[reportAssignmentType]
    """Existing S3 client."""
    key: str | None = None
    """The AWS Access Key."""
    secret: str | None = None
    """The AWS Secret Key."""
    region: str | None = None
    """AWS Region of the bucket."""
    endpoint: str | None = None
    """Custom AWS endpoint."""

    _required_options: ClassVar[list[str]] = ["bucket"]

    def __post_init__(
        self,
        **kwargs: Any,
    ):
        super().__post_init__(**kwargs)

        # leading slash generally does not break anything, but SCAN relies on
        # method that does not work with leading slash. That's why it's
        # stripped, which shouldn't be a problem as boto3 automatically strips
        # it in majority of operations anyway.
        self.path = self.path.strip("/")

        if self.client is None:  # pyright: ignore[reportUnnecessaryComparison]
            self.client = boto3.client(
                "s3",
                aws_access_key_id=self.key,
                aws_secret_access_key=self.secret,
                region_name=self.region,
                endpoint_url=self.endpoint,
            )

        try:
            self.client.head_bucket(Bucket=self.bucket)
        except self.client.exceptions.ClientError as err:
            if self.initialize:
                self.client.create_bucket(Bucket=self.bucket)
            else:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name, f"container {self.bucket} does not exist"
                ) from err


class Reader(fk.Reader):
    """AWS S3 reader."""

    storage: S3Storage
    capabilities: fk.Capability = fk.Capability.STREAM | fk.Capability.LINK_PERMANENT

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        client = self.storage.settings.client
        filepath = self.storage.full_path(data.location)

        try:
            obj: Any = client.get_object(Bucket=self.storage.settings.bucket, Key=filepath)
        except client.exceptions.NoSuchKey as err:
            raise fk.exc.MissingFileError(
                self.storage.settings.name,
                data.location,
            ) from err

        return obj["Body"]

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        name = self.storage.full_path(data.location)
        signed = self.storage.settings.client.generate_presigned_url(
            "get_object", Params={"Bucket": self.storage.settings.bucket, "Key": name}
        )

        return urlunparse(urlparse(signed)._replace(query=None))


class Uploader(fk.Uploader):
    """AWS S3 uploader."""

    storage: S3Storage

    capabilities: fk.Capability = fk.Capability.CREATE | fk.Capability.MULTIPART

    @override
    def upload(self, location: fk.types.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        if not self.storage.settings.override_existing and self.storage.exists(fk.FileData(location), **extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        filepath = self.storage.full_path(location)

        client = self.storage.settings.client

        obj = client.put_object(
            Bucket=self.storage.settings.bucket,
            Key=filepath,
            Body=upload.stream,  # pyright: ignore[reportArgumentType]
            ContentType=upload.content_type,
        )

        filehash = obj["ETag"].strip('"')

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            filehash,
        )

    def _presigned_part(self, key: str, upload_id: str, part: int):
        """Generate a presigned URL for uploading a part."""
        return self.storage.settings.client.generate_presigned_url(
            "upload_part",
            Params={
                "Bucket": self.storage.settings.bucket,
                "Key": key,
                "UploadId": upload_id,
                "PartNumber": part,
            },
        )

    @override
    def multipart_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        client = self.storage.settings.client

        params: CreateMultipartUploadRequestTypeDef = {
            "Bucket": self.storage.settings.bucket,
            "Key": filepath,
        }
        if content_type := extras.get("content_type"):
            params["ContentType"] = content_type

        obj = client.create_multipart_upload(**params)

        return fk.FileData.from_dict(
            extras,
            location=location,
            size=size,
            storage_data={
                "multipart": True,
                "upload_id": obj["UploadId"],
                "uploaded": 0,
                "parts": {},
            },
        )

    @override
    def multipart_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        client = self.storage.settings.client
        filepath = self.storage.full_path(data.location)

        resp = client.list_multipart_uploads(
            Bucket=self.storage.settings.bucket,
            Prefix=filepath,
            MaxUploads=1,
        )

        if "Uploads" in resp:
            client.abort_multipart_upload(
                Bucket=self.storage.settings.bucket,
                Key=filepath,
                UploadId=data.storage_data["upload_id"],
            )
            return True

        return False

    @override
    def multipart_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        client = self.storage.settings.client
        filepath = self.storage.full_path(data.location)

        marker = 0
        parts = {}
        uploaded = 0
        upload_id = data.storage_data.get("upload_id", "")
        try:
            while True:
                resp = client.list_parts(
                    Bucket=self.storage.settings.bucket,
                    Key=filepath,
                    UploadId=upload_id,
                    PartNumberMarker=marker,
                )

                if "Parts" not in resp:
                    break

                for item in resp["Parts"]:
                    if "PartNumber" not in item or "ETag" not in item:
                        continue

                    parts[item["PartNumber"] - 1] = item["ETag"].strip('"')
                    uploaded += item.get("Size", 0)

                marker = resp["NextPartNumberMarker"]
                if not marker:
                    break
        except client.exceptions.NoSuchUpload as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err

        return fk.FileData.from_object(
            data,
            storage_data={"multipart": True, "upload_id": upload_id, "uploaded": uploaded, "parts": parts},
        )

    @override
    def multipart_update(self, data: fk.FileData, upload: fk.Upload, part: int, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        first_byte = data.storage_data["uploaded"]
        last_byte = first_byte + upload.size

        size = data.size

        if last_byte > size:
            raise fk.exc.UploadOutOfBoundError(last_byte, size)

        if last_byte < size and upload.size < 1024 * 1024 * 5:
            msg = "Only the final part can be smaller than 5MiB"
            raise fk.exc.MultipartUploadError(msg)

        resp = self.storage.settings.client.upload_part(
            Bucket=self.storage.settings.bucket,
            Key=filepath,
            UploadId=data.storage_data["upload_id"],
            PartNumber=part + 1,
            Body=upload.stream,  # pyright: ignore[reportArgumentType]
        )

        etag = resp["ETag"].strip('"')
        data.storage_data["uploaded"] = data.storage_data["uploaded"] + upload.size

        data.storage_data["parts"][part] = etag

        return data

    @override
    def multipart_complete(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(data.location)

        if data.size != data.storage_data["uploaded"]:
            raise fk.exc.UploadSizeMismatchError(data.storage_data["uploaded"], data.size)

        result = self.storage.settings.client.complete_multipart_upload(
            Bucket=self.storage.settings.bucket,
            Key=filepath,
            UploadId=data.storage_data["upload_id"],
            MultipartUpload={
                "Parts": [{"PartNumber": int(num) + 1, "ETag": tag} for num, tag in data.storage_data["parts"].items()]
            },
        )

        obj = self.storage.settings.client.get_object(Bucket=self.storage.settings.bucket, Key=result["Key"])

        return fk.FileData(
            fk.types.Location(
                os.path.relpath(
                    result["Key"],
                    self.storage.settings.path,
                )
            ),
            obj["ContentLength"],
            obj["ContentType"],
            obj["ETag"].strip('"'),
        )


class Manager(fk.Manager):
    """AWS S3 manager."""

    storage: S3Storage

    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.ANALYZE
        | fk.Capability.EXISTS
        | fk.Capability.SCAN
        | fk.Capability.MOVE
        | fk.Capability.COPY
        | fk.Capability.SIGNED
    )

    @override
    def signed(self, action: fk.types.SignedAction, duration: int, location: fk.Location, extras: dict[str, Any]):
        """Make an URL for signed action.

        This method uses `generate_presigned_url` internally, which does not
        support content restrictions. If you need to enforce content
        type/size/hash validation, consider `generate_presigned_post`.

        @inherit_doc
        """
        client = self.storage.settings.client
        method = {
            "download": "get_object",
            "upload": "put_object",
            "delete": "delete_object",
        }[action]

        key = self.storage.full_path(location)

        return client.generate_presigned_url(
            ClientMethod=method,
            Params={"Bucket": self.storage.settings.bucket, "Key": key},
            ExpiresIn=duration,
        )

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        client = self.storage.settings.client
        bucket = self.storage.settings.bucket

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        old_key = self.storage.full_path(data.location)
        new_key = self.storage.full_path(location)

        client.copy_object(Bucket=bucket, CopySource={"Bucket": bucket, "Key": old_key}, Key=new_key)

        return self.analyze(location, extras)

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        result = self.copy(location, data, extras)
        self.remove(data, extras)

        return result

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        client = self.storage.settings.client
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        marker = ""

        while True:
            resp = client.list_objects(
                Bucket=self.storage.settings.bucket,
                Marker=marker,
                Prefix=path,
            )
            if "Contents" not in resp:
                break

            for item in resp["Contents"]:
                if "Key" not in item:
                    continue
                key = item["Key"]
                yield os.path.relpath(key, path)

            if "NextMarker" not in resp:
                break

            marker = resp["NextMarker"]

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        client = self.storage.settings.client

        try:
            client.get_object(Bucket=self.storage.settings.bucket, Key=filepath)
        except client.exceptions.NoSuchKey:
            return False

        return True

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        filepath = self.storage.full_path(data.location)
        client = self.storage.settings.client

        try:
            client.head_object(Bucket=self.storage.settings.bucket, Key=filepath)
        except client.exceptions.ClientError:
            return False

        client.delete_object(Bucket=self.storage.settings.bucket, Key=filepath)

        return True

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        filepath = self.storage.full_path(location)
        client = self.storage.settings.client

        try:
            obj = client.get_object(Bucket=self.storage.settings.bucket, Key=filepath)
        except client.exceptions.NoSuchKey as err:
            raise fk.exc.MissingFileError(self.storage, filepath) from err

        return fk.FileData(
            location,
            size=obj["ContentLength"],
            content_type=obj["ContentType"],
            hash=obj["ETag"].strip('"'),
        )


class S3Storage(fk.Storage):
    """AWS S3 storage adapter.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:s3",
        "bucket": "my-bucket,
        "key": "<access key>",
        "secret": "<secret key>",
        "initialize": True,
        ## uncomment following line to use MinIO on port 9000
        # "endpoint": "http://127.0.0.1:9000/",
    }
    storage = fk.make_storage("s3", settings)
    ```

    Note:
    * The `bucket` option is required.
    * If `initialize` is set to `True`, the bucket will be created if it does not exist.
    * If `key` and `secret` are not provided, the default AWS credentials chain
      will be used.
    * The `path` option can be used to set a prefix within the bucket.
    * The `region` option can be used to set the AWS region for the bucket.
    * The `endpoint` option can be used to set a custom endpoint, e.g. for MinIO.
    * The `override_existing` option controls whether existing files can be overwritten.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Default transformers for file_keeper."""

from __future__ import annotations

import logging
import mimetypes
import os
import uuid
from datetime import datetime, timezone
from typing import Any

from file_keeper import BaseData, Upload

FILE_KEEPER_DNS = uuid.UUID("5b762d43-ec0d-3270-a565-8bb44bdaf6cf")

log = logging.getLogger(__name__)


def uuid_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Transform location into random UUID."""
    dir = os.path.split(location)[0]
    return os.path.join(dir, str(uuid.uuid4()))


def uuid_prefix_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Prefix the location with random UUID."""
    dir, file = os.path.split(location)
    return os.path.join(dir, f"{uuid.uuid4()}{file}")


def uuid_with_extension_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Replace location with random UUID, but keep the original extension."""
    dir, file = os.path.split(location)
    ext = os.path.splitext(file)[1]
    return os.path.join(dir, str(uuid.uuid4()) + ext)


def static_uuid_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Transform location into static UUID.

    The same location always transformed into the same UUID. This transformer
    combined with `fix_extension` can be used as an alternative to the
    `safe_relative_path` if you want to avoid nested folders.
    """
    dir = os.path.split(location)[0]
    return os.path.join(dir, str(uuid.uuid5(FILE_KEEPER_DNS, location)))


def datetime_prefix_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Prefix location with current date-timestamp."""
    dir, file = os.path.split(location)
    now = datetime.now(timezone.utc).isoformat()
    return os.path.join(dir, now + file)


def datetime_with_extension_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Replace location with current date-timestamp, but keep the extension."""
    dir, file = os.path.split(location)
    ext = os.path.splitext(file)[1]
    now = datetime.now(timezone.utc).isoformat()
    return os.path.join(dir, now + ext)


def fix_extension_transformer(location: str, upload: Upload | BaseData | None, extras: dict[str, Any]) -> str:
    """Choose extension depending on MIME type of upload.

    When upload is not specified, transformer does nothing.
    """
    if not upload:
        log.debug("Location %s remains unchanged because upload is not specified", location)
        return location

    name = os.path.splitext(location)[0]

    if ext := mimetypes.guess_extension(upload.content_type):
        return name + ext

    log.debug(
        "Location %s remains unchanged because of unexpected upload: %s",
        location,
        upload,
    )
    return location


def safe_relative_path_transformer(location: str, upload: Upload | None, extras: dict[str, Any]) -> str:
    """Remove unsafe segments from path and strip leading slash."""
    return os.path.normpath(location).lstrip("./")



"""Apache Libcloud adapter."""

from __future__ import annotations

import contextlib
import dataclasses
import os
from collections.abc import Iterable
from typing import Any, cast

import requests
from libcloud.base import (
    DriverType,
    get_driver,  # pyright: ignore[reportUnknownVariableType]
)
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Container, StorageDriver
from libcloud.storage.types import ContainerDoesNotExistError, ObjectDoesNotExistError
from typing_extensions import override

import file_keeper as fk

get_driver: Any


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Libcloud settings."""

    provider: str = ""
    key: str = ""
    """Access key of the cloud account."""

    secret: str | None = None
    """Secret key of the cloud account."""

    container_name: str = ""
    """Name of the cloud container."""
    params: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))
    """Additional parameters for cloud provider."""

    public_prefix: str = ""
    """Root URL for containers with public access."""

    driver: StorageDriver = None  # pyright: ignore[reportAssignmentType]
    """Existing storage driver."""
    container: Container = None  # pyright: ignore[reportAssignmentType]
    """Existing container object."""

    def __post_init__(
        self,
        **kwargs: Any,
    ):
        super().__post_init__(**kwargs)

        # leading slash generally does not break anything, but SCAN relies on
        # method that does not work with leading slash. That's why it's
        # stripped, which shouldn't be a problem as libcloud automatically
        # strips it in majority of operations anyway.
        self.path = self.path.strip("/")

        if self.driver is None:  # pyright: ignore[reportUnnecessaryComparison]
            try:
                make_driver = get_driver(DriverType.STORAGE, self.provider)
            except AttributeError as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    str(err),
                ) from err

            self.driver = make_driver(self.key, self.secret, **self.params)

        if self.container is None:  # pyright: ignore[reportUnnecessaryComparison]
            try:
                self.container = self.driver.get_container(self.container_name)

            except ContainerDoesNotExistError as err:
                if self.initialize:
                    self.container = self.driver.create_container(self.container_name)
                else:
                    raise fk.exc.InvalidStorageConfigurationError(
                        self.name, f"container {self.container_name} does not exist"
                    ) from err

            except (LibcloudError, requests.RequestException) as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    str(err),
                ) from err


class Uploader(fk.Uploader):
    """Libcloud uploader."""

    storage: LibCloudStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(self, location: fk.types.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        dest = self.storage.full_path(location)

        if not self.storage.settings.override_existing:
            with contextlib.suppress(ObjectDoesNotExistError):
                self.storage.settings.container.get_object(dest)
                raise fk.exc.ExistingFileError(self.storage, location)

        result = self.storage.settings.container.upload_object_via_stream(
            iter(upload.stream),
            dest,
            extra={"content_type": upload.content_type},
        )

        return fk.FileData(
            location,
            result.size,
            upload.content_type,
            result.hash.strip('"'),
        )


class Reader(fk.Reader):
    """Libcloud reader."""

    storage: LibCloudStorage
    capabilities: fk.Capability = fk.Capability.STREAM | fk.Capability.LINK_PERMANENT

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        location = self.storage.full_path(data.location)

        try:
            obj = self.storage.settings.container.get_object(location)
        except ObjectDoesNotExistError as err:
            raise fk.exc.MissingFileError(
                self.storage,
                data.location,
            ) from err

        return obj.as_stream()

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        location = self.storage.full_path(data.location)
        return os.path.join(self.storage.settings.public_prefix, location)

        # try:
        #     obj = self.storage.settings.container.get_object(location)
        # except ObjectDoesNotExistError as err:
        #     raise fk.exc.MissingFileError(
        #         self.storage,
        #         data.location,
        #     ) from err

        # return self.storage.settings.driver.get_object_cdn_url(obj)


class Manager(fk.Manager):
    """Libcloud manager."""

    storage: LibCloudStorage
    capabilities: fk.Capability = (
        fk.Capability.SCAN | fk.Capability.REMOVE | fk.Capability.EXISTS | fk.Capability.ANALYZE
    )

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        for item in self.storage.settings.container.iterate_objects(prefix=path):
            yield os.path.relpath(item.name, path)

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        location = self.storage.full_path(data.location)

        try:
            obj = self.storage.settings.container.get_object(location)
        except ObjectDoesNotExistError:
            return False
        return self.storage.settings.container.delete_object(obj)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        location = self.storage.full_path(data.location)

        try:
            self.storage.settings.container.get_object(location)
        except ObjectDoesNotExistError:
            return False

        return True

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        fullpath = self.storage.full_path(location)

        try:
            obj = self.storage.settings.container.get_object(fullpath)
        except ObjectDoesNotExistError as err:
            raise fk.exc.MissingFileError(self.storage, location) from err

        content_type: str = obj.extra.get("content_type", fk.FileData.content_type)  # pyright: ignore[reportUnknownVariableType]
        hash = obj.hash.strip('"')
        return fk.FileData(location, obj.size, content_type, hash)


class LibCloudStorage(fk.Storage):
    """Libcloud storage adapter.

    This adapter uses [Apache Libcloud](https://libcloud.apache.org/) to connect to
    various cloud storage providers. It supports any provider implemented in Libcloud,
    such as Amazon S3, Google Cloud Storage, Azure Blob Storage, and many others.

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:libcloud",
        "name": "my_libcloud_storage",
        "provider": "S3",  # or "GOOGLE_STORAGE", "AZURE_BLOBS", etc.
        "key": "<access key>",
        "secret": "<secret key>",
        "container_name": "file-keeper",
        "params": {
            "region": "us-west-1",  # provider-specific parameters
        },
        "public_prefix": "https://my-bucket.s3.amazonaws.com/",  # optional, if provider supports public links
        "path": "uploads/",
        "override_existing": False,
        "initialize": True,
    }

    storage = fk.make_storage("libcloud", settings)
    ```

    Note:
    * Ensure that the `provider` field matches one of the supported Libcloud storage providers.
    * The `params` field can include any additional parameters required by the specific provider.
    * If `public_prefix` is not set, permanent links will not be available.
    * The `initialize` flag allows automatic creation of the container if it does not exist.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    @override
    def compute_capabilities(self) -> fk.Capability:
        cluster = super().compute_capabilities()
        if not self.settings.public_prefix:
            cluster = cluster.exclude(fk.Capability.LINK_PERMANENT)

        return cluster



"""Null adapter."""

from __future__ import annotations

import dataclasses
import logging
from collections.abc import Iterable
from typing import Any

from typing_extensions import override

import file_keeper as fk

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for Null storage."""


class Uploader(fk.Uploader):
    """Null uploader."""

    storage: NullStorage
    capabilities: fk.Capability = fk.Capability.UPLOADER_CAPABILITIES

    @override
    def upload(self, location: fk.Location, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        reader = upload.hashing_reader()
        return fk.FileData(location, hash=reader.get_hash())

    @override
    def multipart_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_dict(extras, size=size, location=location)

    @override
    def multipart_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        return data

    @override
    def multipart_update(self, data: fk.FileData, upload: fk.Upload, part: int, extras: dict[str, Any]) -> fk.FileData:
        return data

    @override
    def multipart_complete(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_object(data)

    @override
    def multipart_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        return False

    @override
    def resumable_start(self, location: fk.Location, size: int, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_dict(extras, location=location, size=size)

    @override
    def resumable_refresh(self, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        return data

    @override
    def resumable_resume(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        return data

    @override
    def resumable_remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        return False


class Manager(fk.Manager):
    """Null manager."""

    storage: NullStorage
    capabilities: fk.Capability = fk.Capability.MANAGER_CAPABILITIES

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        return False

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        return False

    @override
    def compose(self, location: fk.Location, data: Iterable[fk.FileData], extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData(location)

    @override
    def append(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_object(data)

    @override
    def copy(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_object(data, location=location)

    @override
    def move(self, location: fk.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData.from_object(data, location=location)

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        return []

    @override
    def analyze(self, location: fk.Location, extras: dict[str, Any]) -> fk.FileData:
        return fk.FileData(location)

    @override
    def signed(self, action: fk.SignedAction, duration: int, location: fk.Location, extras: dict[str, Any]) -> str:
        return location


class Reader(fk.Reader):
    """Null reader."""

    storage: NullStorage
    capabilities: fk.Capability = fk.Capability.READER_CAPABILITIES

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        return []

    @override
    def range(self, data: fk.FileData, start: int, end: int | None, extras: dict[str, Any]) -> Iterable[bytes]:
        return []

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        return data.location

    @override
    def temporal_link(self, data: fk.FileData, duration: int, extras: dict[str, Any]) -> str:
        return data.location

    @override
    def one_time_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        return data.location


class NullStorage(fk.Storage):
    """Null storage adapter.

    Does not store any files, all operations are no-ops. Useful for testing or
    as a placeholder.

    ```python
    import file_keeper as fk

    settings = {
        "type": "file_keeper:null",
    }

    storage = make_storage("null", settings)
    ```

    Note:
    * Uploading a file will return a `FileData` with the correct hash, but the file
      will not be stored.
    * All manager operations will return default values (e.g., `exists` will always
        return `False`).
    * All reader operations will return empty content or the location as the link.
    """

    settings: Settings

    SettingsFactory = Settings
    UploaderFactory = Uploader
    ReaderFactory = Reader
    ManagerFactory = Manager



"""Type definitions."""

from __future__ import annotations

from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Any, Literal, NewType, Protocol, TypeAlias

if TYPE_CHECKING:
    from .upload import Upload

Location = NewType("Location", str)

LocationTransformer: TypeAlias = Callable[[str, "Upload | None", "dict[str, Any]"], str]

SignedAction = Literal["upload", "download", "delete"]


class PReadable(Protocol):
    """Readable object."""

    def read(self, size: Any = ..., /) -> bytes:
        """Read up to `size` bytes. If the argument is negative or omitted, read until EOF.

        Return an empty bytes object on EOF.
        """
        ...


class PStream(PReadable, Protocol):
    """Readable stream."""

    def __iter__(self) -> Iterator[bytes]:
        """Return an iterator over the stream."""
        ...


class PSeekableStream(PStream, Protocol):
    """Stream that supports `seek` operation."""

    def tell(self) -> int:
        """Get the current position of the pointer."""
        ...

    def seek(self, offset: int, whence: int = 0) -> int:
        """Move pointer to the specified position."""
        ...


__all__ = [
    "PStream",
    "PSeekableStream",
]



"""Filebin adapter."""

from __future__ import annotations

import base64
import dataclasses
from collections.abc import Iterable
from typing import IO, Any, ClassVar

import requests
from typing_extensions import override

import file_keeper as fk

API_URL = "https://filebin.net"


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Filebin settings."""

    timeout: int = 10
    bin: str = ""

    _required_options: ClassVar[list[str]] = ["bin"]


class Uploader(fk.Uploader):
    """Filebin uploader."""

    storage: FilebinStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        resp = requests.post(
            f"{API_URL}/{self.storage.settings.bin}/{location}",
            data=upload.stream,
            timeout=self.storage.settings.timeout,
        )
        if not resp.ok:
            raise fk.exc.UploadError(resp.content)

        info: dict[str, Any] = resp.json()["file"]
        return fk.FileData(
            info["filename"],
            upload.size,
            upload.content_type,
            base64.decodebytes(info["md5"].encode()).decode(),
        )


class Reader(fk.Reader):
    """Filebin reader."""

    storage: FilebinStorage
    capabilities: fk.Capability = fk.Capability.STREAM | fk.Capability.LINK_PERMANENT

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> IO[bytes]:
        resp = requests.get(
            f"{API_URL}/{self.storage.settings.bin}/{data.location}",
            timeout=self.storage.settings.timeout,
            stream=True,
            headers={"accept": "*/*"},
        )
        if verified := resp.cookies.get("verified"):
            resp = requests.get(
                f"{API_URL}/{self.storage.settings.bin}/{data.location}",
                cookies={"verified": verified},
                timeout=self.storage.settings.timeout,
                stream=True,
                headers={"accept": "*/*"},
            )

        return resp.raw  # pyright: ignore[reportReturnType]

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        return f"{API_URL}/{self.storage.settings.bin}/{data.location}"


class Manager(fk.Manager):
    """Filebin adapter."""

    storage: FilebinStorage
    capabilities: fk.Capability = fk.Capability.REMOVE | fk.Capability.SCAN | fk.Capability.ANALYZE

    @override
    def remove(
        self,
        data: fk.FileData,
        extras: dict[str, Any],
    ) -> bool:
        requests.delete(
            f"{API_URL}/{self.storage.settings.bin}/{data.location}",
            timeout=self.storage.settings.timeout,
        )
        return True

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        resp = requests.get(
            f"{API_URL}/{self.storage.settings.bin}",
            headers={"accept": "application/json"},
            timeout=self.storage.settings.timeout,
        )

        for record in resp.json()["files"]:
            yield record["filename"]

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        resp = requests.get(
            f"{API_URL}/{self.storage.settings.bin}",
            headers={"accept": "application/json"},
            timeout=self.storage.settings.timeout,
        )
        for record in resp.json()["files"]:
            if record["filename"] == location:
                return fk.FileData(
                    record["filename"],
                    record["size"],
                    record["content-type"],
                    base64.decodebytes(record["md5"].encode()).decode(),
                )

        raise fk.exc.MissingFileError(self.storage, location)


class FilebinStorage(fk.Storage):
    """Filebin adapter."""

    hidden = True

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import copy
import dataclasses
import operator
from collections.abc import Callable
from typing import Any, ClassVar, cast

from typing_extensions import TypeVar

from . import types

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class BaseData:
    """Base class for file details."""

    location: types.Location
    size: int = 0
    content_type: str = ""
    hash: str = ""
    storage_data: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))

    _plain_keys: ClassVar[list[str]] = ["location", "size", "content_type", "hash"]
    _complex_keys: ClassVar[list[str]] = ["storage_data"]

    @classmethod
    def from_string(cls, location: str | types.Location):
        """Create data object from location string."""
        return cls(types.Location(location))

    @classmethod
    def from_dict(cls, record: dict[str, Any], **overrides: Any):
        """Transform dictionary into data object."""
        return cls._from(record, operator.getitem, operator.contains, overrides)

    @classmethod
    def from_object(cls, obj: Any, **overrides: Any):
        """Copy data details from another object."""
        return cls._from(obj, getattr, hasattr, overrides)

    @classmethod
    def _from(
        cls,
        source: Any,
        getter: Callable[[Any, str], Any],
        checker: Callable[[Any, str], bool],
        overrides: dict[str, Any],
    ):
        data = {}
        for key in cls._plain_keys:
            if key in overrides:
                data[key] = overrides[key]
            elif checker(source, key):
                data[key] = getter(source, key)

        for key in cls._complex_keys:
            if key in overrides:
                data[key] = overrides[key]
            elif checker(source, key):
                data[key] = copy.deepcopy(getter(source, key))

        return cls(**data)

    def into_object(self, obj: T) -> T:
        """Copy data attributes into another object."""
        for key in self._plain_keys:
            setattr(obj, key, getattr(self, key))

        for key in self._complex_keys:
            setattr(obj, key, copy.deepcopy(getattr(self, key)))

        return obj

    def as_dict(self):
        """Return data as dictionary."""
        return dataclasses.asdict(self)


@dataclasses.dataclass(frozen=True)
class FileData(BaseData):
    """Information required by storage to operate the file.

    Args:
        location: filepath, filename or any other type of unique identifier
        size: size of the file in bytes
        content_type: MIMEtype of the file
        hash: checksum of the file
        storage_data: additional details set by storage adapter

    Example:
        ```
        FileData(
            "local/path.txt",
            123,
            "text/plain",
            md5_of_content,
        )
        ```
    """

    content_type: str = "application/octet-stream"



"""OpenDAL adapter."""

from __future__ import annotations

import dataclasses
import io
import os
from collections.abc import Iterable
from typing import Any, cast

import magic
import opendal
from typing_extensions import override

import file_keeper as fk


class FileStream:
    """Wrapper for stream returned by OpenDAL."""

    file: opendal.File

    def __init__(self, file: opendal.File):
        self.file = file

    def __iter__(self):
        while chunk := self.file.read(io.DEFAULT_BUFFER_SIZE):
            if isinstance(chunk, memoryview):  # pyright: ignore[reportUnnecessaryIsInstance]
                yield chunk.tobytes()
            else:
                yield chunk


@dataclasses.dataclass()
class Settings(fk.Settings):
    """OpenDAL settings."""

    params: dict[str, Any] = cast("dict[str, Any]", dataclasses.field(default_factory=dict))
    """Parameters for OpenDAL operator initialization."""
    scheme: str = ""
    """Name of OpenDAL operator's scheme."""
    operator: opendal.Operator = None  # pyright: ignore[reportAssignmentType]
    """Existing OpenDAL operator."""

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        # operator has separate parameter for the root and `path` must be
        # relative to it
        self.path = self.path.lstrip("/")

        if self.operator is None:  # pyright: ignore[reportUnnecessaryComparison]
            if not self.scheme:
                raise fk.exc.MissingStorageConfigurationError(self.name, "scheme")

            try:
                self.operator = opendal.Operator(self.scheme, **self.params)
            except opendal.exceptions.ConfigInvalid as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    self.name,
                    str(err),
                ) from err


class Uploader(fk.Uploader):
    """OpenDAL uploader."""

    storage: OpenDalStorage
    capabilities: fk.Capability = fk.Capability.CREATE

    @override
    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Upload file to computed location.

        Schemas working with filesystem-like paths assume that location is
        relative the configured `path`. The location is not sanitized and can
        lead outside the configured `path`. Consider using combination of
        `storage.prepare_location` with `settings.location_transformers` that
        sanitizes the path, like `safe_relative_path`.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            LocationError: unallowed usage of subdirectory

        Returns:
            New file data

        """
        op = self.storage.settings.operator
        dest = self.storage.full_path(location)
        subpath = os.path.dirname(dest)

        if subpath and not op.capability().create_dir:
            raise fk.exc.LocationError(self.storage, subpath)

        try:
            op.stat(dest)
        except opendal.exceptions.NotFound:
            pass
        else:
            if not self.storage.settings.override_existing:
                raise fk.exc.ExistingFileError(self.storage, location)

        if subpath:
            # opendal reuquires `/` as a last character
            op.create_dir(subpath.rstrip("/") + "/")

        reader = upload.hashing_reader()
        with op.open(dest, "wb") as fobj:
            for chunk in reader:
                fobj.write(chunk)

        return fk.FileData(
            location,
            reader.position,
            upload.content_type,
            reader.get_hash(),
        )


class Reader(fk.Reader):
    """OpenDAL reader."""

    storage: OpenDalStorage
    capabilities: fk.Capability = fk.Capability.STREAM

    @override
    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        location = self.storage.full_path(data.location)

        try:
            content = self.storage.settings.operator.open(location, "rb")
        except opendal.exceptions.NotFound as err:
            raise fk.exc.MissingFileError(self.storage, data.location) from err

        return FileStream(content)


class Manager(fk.Manager):
    """OpenDAL manager."""

    storage: OpenDalStorage
    capabilities: fk.Capability = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.ANALYZE
        | fk.Capability.APPEND
    )

    @override
    def copy(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        op.copy(src_location, dest_location)

        return fk.FileData.from_object(data, location=location)

    @override
    def move(self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]) -> fk.FileData:
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(fk.FileData(location), extras):
            raise fk.exc.ExistingFileError(self.storage, location)

        src_location = self.storage.full_path(data.location)
        dest_location = self.storage.full_path(location)

        op.rename(src_location, dest_location)

        return fk.FileData.from_object(data, location=location)

    @override
    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        location = self.storage.full_path(data.location)

        try:
            self.storage.settings.operator.stat(location)
        except opendal.exceptions.NotFound:
            return False

        return True

    @override
    def analyze(self, location: fk.types.Location, extras: dict[str, Any]) -> fk.FileData:
        stream = self.storage.stream(fk.FileData(location))

        reader = fk.HashingReader(fk.IterableBytesReader(stream))
        content_type = magic.from_buffer(next(reader, b""), True)
        reader.exhaust()

        return fk.FileData(
            location,
            size=reader.position,
            content_type=content_type,
            hash=reader.get_hash(),
        )

    @override
    def remove(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        op = self.storage.settings.operator
        location = self.storage.full_path(data.location)

        try:
            op.stat(location)
        except opendal.exceptions.NotFound:
            return False

        op.delete(location)
        return True

    @override
    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        path = self.storage.settings.path
        # do not add slash when path empty, because it will change it from
        # "current directory" to the "root directory"
        if path:
            path = path.rstrip("/") + "/"

        for entry in self.storage.settings.operator.scan(path):
            stat = self.storage.settings.operator.stat(entry.path)

            if opendal.EntryMode.is_file(stat.mode):
                yield os.path.relpath(entry.path, path)

    @override
    def append(self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]) -> fk.FileData:
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        location = self.storage.full_path(data.location)

        op.write(location, upload.stream.read(), append=True)

        return self.analyze(data.location, extras)


class OpenDalStorage(fk.Storage):
    """OpenDAL storage.

    OpenDAL (https://opendal.apache.org/) is a multi-protocol data access library
    with unified API. It supports a variety of backends, including local filesystem,
    S3-compatible object storage, Azure Blob Storage, Google Cloud Storage, HDFS,
    and many others.

    To configure OpenDAL storage, you need to specify the `scheme` and any
    necessary `params` for the chosen backend. Alternatively, you can provide an
    existing `operator` instance.

    Supported schemes and their parameters can be found in the
    [OpenDAL documentation](https://opendal.apache.org/core/#services).

    Example configuration:

    ```py
    import file_keeper as fk

    settings = {
        "type": "file_keeper:opendal",
        "scheme": "s3",
        "params": {
            "bucket": "file-keeper",
            "access_key_id": "minioadmin",
            "secret_access_key": "minioadmin",
            "endpoint": "http://127.0.0.1:9000",
            "region": "auto",
        },
        "path": "uploads/",
        "override_existing": False,
    }

    storage = fk.make_storage("opendal", settings)
    ```

    Note:
    * The `path` parameter is used as a prefix for all file operations.
    * Ensure that the OpenDAL library is installed in your environment.
    """

    settings: Settings
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    @override
    def compute_capabilities(self) -> fk.Capability:
        cluster = super().compute_capabilities()
        capabilities = self.settings.operator.capability()

        if not capabilities.stat:
            cluster = cluster.exclude(fk.Capability.EXISTS | fk.Capability.ANALYZE)

        if not capabilities.delete:
            cluster = cluster.exclude(fk.Capability.REMOVE)

        if not capabilities.list or not capabilities.stat:
            cluster = cluster.exclude(fk.Capability.SCAN)

        if not capabilities.write:
            cluster = cluster.exclude(fk.Capability.CREATE)

        if not capabilities.read:
            cluster = cluster.exclude(fk.Capability.STREAM)

        if not capabilities.rename:
            cluster = cluster.exclude(fk.Capability.MOVE)

        if not capabilities.copy:
            cluster = cluster.exclude(fk.Capability.COPY)

        if not capabilities.write_can_append:
            cluster = cluster.exclude(fk.Capability.APPEND)

        return cluster
